// MurderSpot - a free game based on Clue/Cluedo
// Copyright 2005-2009 Christopher Elsby <chrise@chrise.me.uk>
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "global.h"

#include "board.h"
#include <cstdlib>
#include <map>
#include <wx/defs.h>
#include <wx/string.h>
#include <wx/textfile.h>
#include <wx/log.h>
#include <wx/image.h>
//#include <wx/filename.h>

Board::Board() : imageWidth(0), imageHeight(0), offsetw(0), offsetr(0), ncards(0), cards(NULL), locations(NULL),
                 ntypes(0), types(NULL), w(0), h(0), board(NULL), ndice(0), dice(NULL),
                 stepresults(NULL), stepresultsadj(NULL), nturnconfig(0), turnconfig(NULL)
{
    for(wxUint16 n = 0; n < 256; ++n)
        turnmap[n] = 255;
}

void Board::unload()
{
    if(cards != NULL)
        delete[] cards;
    if(types != NULL)
        delete[] types;
    if(board != NULL)
        delete[] board;
    if(dice != NULL)
        delete[] dice;
    if(stepresults != NULL)
        delete[] stepresults;
    if(stepresultsadj != NULL)
        delete[] stepresultsadj;
    if(locations != NULL)
        delete[] locations;
    if(turnconfig != NULL)
        delete[] turnconfig;
    offsetw = offsetr = ncards = ntypes = w = h = ndice = /*nmoverules =*/ nturnconfig = 0;
    cards = NULL;
    types = board = dice = NULL;
    locations = NULL;
    stepresults = NULL;
    stepresultsadj = NULL;
    turnconfig = NULL;
    for(wxUint16 n = 0; n < 256; ++n)
        turnmap[n] = 255;
}

bool Board::init(wxUint8 width, wxUint8 height, wxUint8 suspectcount, wxUint8 weaponcount, wxUint8 roomcount, wxUint8 suggestturns, wxUint8 accuseturns)
{
    int i;
    offsetw = suspectcount;
    offsetr = offsetw + weaponcount;
    ncards = offsetr + roomcount;
    if(cards != NULL)
        delete[] cards;
    cards = new wxString[ncards];
    w = width;
    h = height;
    ntypes = roomcount + 1;
    if(types != NULL)
        delete[] types;
    types = new wxUint8[ntypes];
    for(i = 0; i < ntypes; ++i)
        types[i] = i + offsetr;
    if(board != NULL)
        delete[] board;
    board = new wxUint8[w * (wxUint16) h];
    for(i = 0; i < w * (wxUint16) h; ++i)
        board[i] = roomcount;
    if(dice != NULL)
    {
        delete[] dice;
        dice = NULL;
    }
    if(stepresults != NULL)
    {
        delete[] stepresults;
        stepresults = NULL;
    }
    if(stepresultsadj != NULL)
    {
        delete[] stepresultsadj;
        stepresultsadj = NULL;
    }
    if(locations != NULL)
        delete[] locations;
    locations = new wxUint16[suspectcount * 2];
    for(i = 0; i < suspectcount * 2; ++i)
        locations[i] = (wxUint16) -1;
    if(turnconfig != NULL)
    {
        delete[] turnconfig;
        turnconfig = NULL;
    }
    turnmap[63] = suggestturns;
    turnmap[33] = accuseturns;
    return TRUE;
}

inline static wxString safeNextLine(wxTextFile &data) {if(data.Eof()) return wxEmptyString; return data.GetNextLine();}

wxString Board::getTitle(const wxString &file)
{   
    wxTextFile data;
    if(!data.Open(file))
    {   
        wxLogError(wxT("Failed to open file %s."), file.c_str());
        return wxEmptyString;
    }
    if(data.GetFirstLine().MakeLower() != wxT("murderspot data format 1") && data.GetFirstLine().MakeLower() != wxT("cluedome data format 1"))
    {   
        wxLogError(wxT("Unrecognised file format."));
        return wxEmptyString;
    }
    while(!data.Eof())
    {   
        if(safeNextLine(data).MakeLower() == wxT("title:"))
            return safeNextLine(data);
    }
    return file;
}

void Board::setImageFile(const wxString &imgfile, wxUint8 imgw, wxUint8 imgh)
{
    if(imgfile.empty())
    {
        imageFile.clear();
        imgw = 0;
        imgh = 0;
    }
    else
    {
#if 0
        wxFileName filename(imgfile);
        if(filename.IsOk() && filename.FileExists())
        {
            filename.MakeAbsolute();
            imageFile = filename.GetFullPath();
        }
        else
#endif
        {
            imageFile = imgfile;
        }
        imageWidth = imgw;
        imageHeight = imgh;
    }
}

bool Board::load(const wxString &file)
{
    wxTextFile data;
    if(!data.Open(file))
    {
        wxLogError(wxT("Failed to open file %s."), file.c_str());
        return FALSE;
    }
    if(data.GetFirstLine().MakeLower() != wxT("murderspot data format 1") && data.GetFirstLine().MakeLower() != wxT("cluedome data format 1"))
        return FALSE;
    // Map from board character to square type index.
    // Indexed by character, value is square type + 1.
    wxUint8 typemap[256] = {0};
    // Counters.
    wxUint8 x, y;
    wxUint32 i;
    // Count stuff and generate the type map.
    wxString line;
    for(line = data.GetFirstLine(); !data.Eof(); line = safeNextLine(data))
    {
        line.MakeLower();
        if(line == wxT("card names:"))
        {
            sname = safeNextLine(data);
            y = 0;
            while(y < 255 && !safeNextLine(data).empty())
                ++y;
            if(y != offsetw)
            {
                offsetw = y;
                if(locations != NULL)
                {
                    delete[] locations;
                    locations = NULL;
                }
                if(cards != NULL)
                {
                    delete[] cards;
                    cards = NULL;
                }
            }
            if(locations == NULL)
                locations = new wxUint16[offsetw * 2];
            wname = safeNextLine(data);
            while(y < 255 && !safeNextLine(data).empty())
                ++y;
            if(y != offsetr)
            {
                offsetr = y;
                if(cards != NULL)
                {
                    delete[] cards;
                    cards = NULL;
                }
            }
            rname = safeNextLine(data);
            while(y < 255 && !safeNextLine(data).empty())
                ++y;
            if(y != ncards)
            {
                ncards = y;
                if(cards != NULL)
                {
                    delete[] cards;
                    cards = NULL;
                }
            }
            if(cards == NULL)
                cards = new wxString[ncards];
        }
        else if(line == wxT("dice:"))
        {
            ndice = 0;
            while(ndice < 255 && !safeNextLine(data).empty())
                ++ndice;
            if(dice != NULL)
                delete[] dice;
            dice = new wxUint8[ndice];
        }
        else if(line == wxT("points:"))
        {
            nturnconfig = 0;
            while(!(line = safeNextLine(data)).empty())
            {
                if(wxSscanf(line.substr(1).c_str(), wxT("%u"), &i) == 1)
                {
                    if(i > 254)
                    {
                        wxLogError(wxT("Symbol %c has %u turns assigned to it, but the maximum is 254."), line.at(0), i);
                        unload();
                        return false;
                    }
                    turnmap[(wxUint8) line.at(0)] = i;
                }
                else
                    turnmap[(wxUint8) line.at(0)] = 255;
                if(line.find(wxT('~'), 1) < line.size())
                    ++nturnconfig;
            }
            if(turnconfig != NULL)
                delete[] turnconfig;
            if(nturnconfig > 0)
                turnconfig = new wxString[nturnconfig];
        }
#if 0
        else if(line == wxT("movement:"))
        {
            nmoverules = 0;
            while(nmoverules < 255 && !(line = safeNextLine(data)).empty())
                ++nmoverules;
            if(moverules != NULL)
                delete[] moverules;
            moverules = new MoveRule[nmoverules];
        }
#endif
        else if(line == wxT("board:"))
        {
            w = h = ntypes = 0;
            while(ntypes < 255 && !(line = safeNextLine(data)).empty())
            {
                if(++h == 0)
                {
                    wxLogError(wxT("Board has more than 255 rows."));
                }
                else if(w == 0)
                {
                    if(line.size() > 255)
                    {
                        wxLogError(wxT("Board has width %i, but the maximum allowed is 255."), line.size());
                        unload();
                        return FALSE;
                    }
                    w = line.size();
                }
                else if(w != line.size())
                {
                    wxLogError(wxT("Row %i of the board has width %i, but row 1 has width %i."), h, line.size(), w);
                    unload();
                    return FALSE;
                }
                for(x = 0; x < w; ++x)
                {
                    if(typemap[(wxUint8) line.at(x)] == 0)
                    {
                        if(ntypes == 255)
                        {
                            wxLogError(wxT("More than 255 square types."));
                            unload();
                        }
                        typemap[(wxUint8) line.at(x)] = ++ntypes; // Yes, this is meant to start at 1.
                    }
                }
            }
            if(board != NULL)
                delete[] board;
            board = new wxUint8[w * h];
            if(types != NULL)
                delete[] types;
            types = new wxUint8[ntypes];
            for(y = 0; y < ntypes; ++y)
                types[y] = 255;
            if(stepresults != NULL)
                delete[] stepresults;
            stepresults = new StepResult[ntypes * ntypes];
            if(stepresultsadj != NULL)
                delete[] stepresultsadj;
            stepresultsadj = new StepResult[ntypes * ntypes];
        }
    }
    // Check that there is at least 1 card of each category.
    if(offsetw == 0 || offsetr == offsetw || ncards == offsetr)
    {
        wxLogError(wxT("Either the card names section was omitted from the board file, or there was not at least 1 card of each type."));
        return FALSE;
    }
    // Check that the board is not empty.
    if(ntypes == 0)
    {
        wxLogError(wxT("Either the board section was omitted from the board file, or it was empty."));
        return FALSE;
    }
    // Read the actual data.
    wxUint8 starts[offsetw]; // Starting square types (index is suspect card, value is square type).
    for(y = 0; y < offsetw; ++y)
        starts[y] = 255;
    for(line = data.GetFirstLine(); !data.Eof(); line = safeNextLine(data))
    {
        line.MakeLower();
        if(line == wxT("title:"))
        {
            // Board title.
            title = safeNextLine(data);
        }
        else if(line == wxT("image:"))
        {
            // Image title.
            safeNextLine(data);
            // Image filename.
            imageFile = safeNextLine(data);
            // Image cell width.
            line = safeNextLine(data);
            if(wxSscanf(line.c_str(), wxT("%u"), &i) == 1 && i < 256)
                imageWidth = i;
            else
                imageWidth = 0;
            // Image cell height.
            line = safeNextLine(data);
            if(wxSscanf(line.c_str(), wxT("%u"), &i) == 1 && i < 256)
                imageHeight = i;
            else
                imageHeight = 0;
            setImageFile(imageFile, imageWidth, imageHeight);
        }
        else if(line == wxT("dice:"))
        {
            for(y = 0; y < ndice; ++y)
            {
                line = safeNextLine(data);
                if(wxSscanf(line.c_str(), wxT("%u"), &i) == 1 && i < 255)
                    dice[y] = i;
                else
                    dice[y] = 0;
            }
        }
        else if(line == wxT("points:"))
        {
            y = 0;
            while(y < nturnconfig)
            {
                line = safeNextLine(data);
                i = line.find(wxT('~'), 1);
                if(i < line.size())
                {
                    do
                        ++i;
                    while(i < line.size() && (line.at(i) == ' ' || line.at(i) == '\t'));
                    if(i < line.size())
                    {
                        turnconfig[y] = line.substr(0, 1) + line.substr(i);
                        ++y;
                    }
                    else
                    {
                        --nturnconfig;
                    }
                }
            }
        }
        else if(line == wxT("card names:"))
        {
            safeNextLine(data); // Skip the suspects group name.
            for(y = 0; y < ncards; ++y)
            {
                if(y == offsetw || y == offsetr)
                {
                    safeNextLine(data); // Skip the blank line.
                    safeNextLine(data); // Skip the group name.
                }
                cards[y] = safeNextLine(data);
            }
        }
        else if(line == wxT("starting squares:"))
        {
            for(y = 0; y < offsetw; ++y)
            {
                line = safeNextLine(data);
                for(x = 0; x < line.size(); ++x)
                {
                    if(typemap[(wxUint8) line.at(x)] - 1 < ntypes)
                    {
                        starts[y] = typemap[(wxUint8) line.at(x)] - 1;
                        types[typemap[(wxUint8) line.at(x)] - 1] = y;
                    }
                    else
                    {
                        starts[y] = 255;
                    }
                }
            }
        }
        else if(line == wxT("room squares:"))
        {
            for(y = offsetr; y < ncards; ++y)
            {
                line = safeNextLine(data);
                for(x = 0; x < line.size(); ++x)
                {
                    if(typemap[(wxUint8) line.at(x)] - 1 < ntypes)
                        types[typemap[(wxUint8) line.at(x)] - 1] = y;
                }
            }
        }
        else if(line == wxT("board:"))
        {
            for(y = 0; y < h; ++y)
            {
                line = safeNextLine(data);
                for(x = 0; x < w; ++x)
                    board[y * w + x] = typemap[(wxUint8) line.at(x)] - 1;
            }
        }
        else if(line == wxT("movement:"))
        {
            if(stepresults != NULL && stepresultsadj != NULL)
            {
                for(x = 0; x < ntypes; ++x)
                {
                    for(y = 0; y < ntypes; ++y)
                    {
                        i = x + ntypes * y;
                        stepresultsadj[i].turnkey = 255;
                        stepresultsadj[i].warptype = 255;
                        stepresults[i].turnkey = 255;
                        stepresults[i].warptype = 255;
                    }
                }
                wxUint8 fromtypes[ntypes];
                wxUint8 nfromtypes;
                wxUint8 totypes[ntypes];
                wxUint8 ntotypes;
                while(!(line = safeNextLine(data)).empty())
                {
                    nfromtypes = 0;
                    x = 3;
                    while(x < line.size())
                    {
                        y = typemap[(wxUint8) line.at(x)] - 1;
                        ++x;
                        if(y == 255)
                            break;
                        fromtypes[nfromtypes] = y;
                        ++nfromtypes;
                    }
                    if(nfromtypes == 0)
                    {
                        while(nfromtypes < ntypes)
                        {
                            fromtypes[nfromtypes] = nfromtypes;
                            ++nfromtypes;
                        }
                    }
                    ntotypes = 0;
                    while(x < line.size())
                    {
                        y = typemap[(wxUint8) line.at(x)] - 1;
                        ++x;
                        if(y == 255)
                            break;
                        totypes[ntotypes] = y;
                        ++ntotypes;
                    }
                    if(ntotypes == 0)
                    {
                        while(ntotypes < ntypes)
                        {
                            totypes[ntotypes] = ntotypes;
                            ++ntotypes;
                        }
                    }
                    for(x = 0; x < nfromtypes; ++x)
                    {
                        for(y = 0; y < ntotypes; ++y)
                        {
                            i = fromtypes[x] + ntypes * totypes[y];
                            stepresultsadj[i].turnkey = (wxUint8) line.at(0);
                            stepresultsadj[i].warptype = line.size() > 1 ? typemap[(wxUint8) line.at(1)] - 1 : 255;
                            if(line.size() > 2 && typemap[(wxUint8) line.at(2)] != 0)
                                stepresults[i] = stepresultsadj[i];
                        }
                    }
                }
            }
#if 0 // Old version
            for(y = 0; y < nmoverules; ++y)
            {
                line = safeNextLine(data);
                // First character indicates movement points needed.
                moverules[y].movepoints = (wxUint8) line.at(0);
                // Second character indicates type to warp to.
                moverules[y].warp = line.size() > 1 ? typemap[(wxUint8) line.at(1)] - 1 : 255;
                // Third character indicates whether the source and target squares must be adjacent.
                moverules[y].needadj = line.size() < 3 || typemap[(wxUint8) line.at(2)] == 0;
                if(line.size() > 3)
                {
                    // Simply read in the rest of the line.
                    // Valid characters get mapped to the appropriate types.
                    // Invalid characters get mapped to 255.
                    // The size of the array is: length of line - 3 initial characters + 2 terminating 255s.
                    moverules[y].types = new wxUint8[line.size() - 1];
                    for(x = 3; x < line.size(); ++x)
                        moverules[y].types[x - 3] = typemap[(wxUint8) line.at(x)] - 1;
                    // 255s are added at the end to make sure all groups are terminated
                    // correctly regardless of whether they are in the board file.
                    moverules[y].types[line.size() - 1] = 255;
                    moverules[y].types[line.size() - 2] = 255;
                }
                else
                {
                    // Default rule.
                    moverules[y].types = NULL;
                }
            }
#endif
        }
    }
    // Initialise suspect positions.
    for(y = 0; y < offsetw; ++y)
    {
        i = randSquare(starts[y]);
        if(i < (wxUint16) (w * h))
            locations[y] = locations[y + offsetw] = i;
    }
    return TRUE;
}

bool Board::setRoomAt(wxUint8 x, wxUint8 y, wxUint8 card)
{
    if(board == NULL || x >= w || y >= h)
        return FALSE;
    bool noroom = card < offsetr || card >= ncards;
    for(wxUint8 type = 0; type < ntypes; ++type)
    {
        if(noroom ? types[type] < offsetr || types[type] >= ncards : types[type] == card)
        {
            board[x + w * y] = type;
            return TRUE;
        }
    }
    return FALSE;
}

wxUint8 Board::getRoll() const
{
    if(ndice < 1)
        return 0;
    wxUint8 roll = dice[0];
    for(wxUint8 n = 1; n < ndice; ++n)
    {
        if(dice[n] > 0)
            roll += std::rand() / (RAND_MAX / dice[n] + 1); // Uses high-order bits of rand() - low-order bits are allegedly unreliable.
    }
    return roll;
}

wxUint8 Board::getMinRoll() const
{
    if(ndice < 1)
        return 0;
    return dice[0];
}

wxUint8 Board::getMaxRoll() const
{
    if(ndice < 1)
        return 0;
    wxUint8 roll = dice[0];
    for(wxUint8 n = 1; n < ndice; ++n)
        roll += dice[n] - 1;
    return roll;
}

wxString Board::getDiceString() const
{
    if(ndice < 1)
        return wxEmptyString;
    wxString str;
    str << (int) dice[0];
    if(ndice < 2)
        return str;
    if(dice[1] > 1)
        str << wxT("-") << ((int) dice[1] + (int) dice[0] - 1);
    for(wxUint8 n = 2; n < ndice; ++n)
    {
        if(dice[n] > 1)
            str << wxT(" + 0-") << ((int) dice[n] - 1);
    }
    return str;
}

bool Board::setDiceString(wxString str)
{
    size_t n;
    size_t i, i2;
    long min, max;
    i = 0;
    while(i < str.size())
    {
        if(str[i] == wxT(' '))
        {
            str.erase(i, 1);
            continue;
        }
        else if(str[i] != wxT('+') && str[i] != wxT('-') && (str[i] < wxT('0') || str[i] > wxT('9')))
        {
            return FALSE;
        }
        ++i;
    }
    wxUint8 *newdice = new wxUint8[255];
    for(n = 0; n < 255; ++n)
        newdice[n] = 0;
    n = 0;
    while(1)
    {
//printf("Loop\n");
        if(n >= 255)
            goto Board_setDiceString_fail;
        i = str.find(wxT('+'));
        i2 = str.find(wxT('-'));
        if(i2 < i)
        {
            if(i < str.size())
            {
                if(!wxString(str.substr(0, i2)).ToLong(&min) || !wxString(str.substr(i2 + 1, i - i2 - 1)).ToLong(&max))
{//printf("Failure 1\n");
                    goto Board_setDiceString_fail;
}
            }
            else
            {
                if(!wxString(str.substr(0, i2)).ToLong(&min) || !wxString(str.substr(i2 + 1)).ToLong(&max))
{//printf("Failure 2\n");
                    goto Board_setDiceString_fail;
}
            }
        }
        else
        {
            if(!wxString(str.substr(0, i)).ToLong(&min))
{//printf("Failure 3\n");
                goto Board_setDiceString_fail;
}
            max = min;
        }
        if(n < 1)
            n = 1;
        if(min + newdice[0] < 0 || min + newdice[0] > 255 || max - min + 1 < 0 || max - min + 1 > 255)
{//printf("Failure 4\n");
            goto Board_setDiceString_fail;
}
        newdice[0] += min;
        newdice[n] = max - min + 1;
        if(newdice[n] > 1)
            ++n;
        if(i >= str.size())
            break;
        str.erase(0, i + 1);
    }
    if(n < 1)
{//printf("Failure 5\n");
        goto Board_setDiceString_fail;
}
    if(dice != NULL)
        delete [] dice;
    dice = newdice;
    ndice = n;
    return TRUE;
Board_setDiceString_fail:
    delete [] newdice;
    return FALSE;
}

wxUint8 Board::getMoveTurns(wxUint8 fromX, wxUint8 fromY, wxUint8 toX, wxUint8 toY, wxUint8 max, wxUint8 *finalX, wxUint8 *finalY) const
{
    // We find the best route using Dijkstra's algorithm.
    // This is based on an implementation which I saw in the Wikipedia article on it.
    // It would be rather complicated to deal with warps across multiple steps, so we only allow warp for the final step.
    // To facilitate this warp policy, we actually perform the search backwards.
    // To decrease search times, we do not give every board square a node in the graph on which we are applying the algorithm.
    // Instead, we only use one square of each type as a node, except for those types for which position is important (i.e. there
    // is a difference between its stepresults and stepresultsadj values).
    // FIXME: We should not allow paths that involve entering and then leaving a room.
    // The current code will only work if entering a room always leads to a warp.
    // TODO: Allow warp moves as intermediates; if this is done, the above room checking absolutely must be done too.

    // Standard values of finalX and finalY.
    if(finalX != NULL)
        *finalX = toX;
    if(finalY != NULL)
        *finalY = toY;

    // Basic sanity checks.
    if(fromX == toX && fromY == toY)
        return 0;
    if(fromX >= w || fromY >= h || toX >= w || toY >= h || board == NULL || stepresults == NULL || stepresultsadj == NULL)
        return 255;
    if(max > 254)
        max = 254;

    // Variables for general use.
    wxUint8 type, type2;
    wxUint16 square, square2;
    wxUint16 iunvisited, iunvisited2;
    wxUint16 distance, distance2; // wxUint16 is used so that we can safely add distances.
    const StepResult *stepresult;

    // The total number of squares on the board.
    wxUint16 nsquares = w * h;

    // The starting and finishing squares.
    wxUint16 fromSquare = fromX + w * fromY;
    wxUint16 toSquare = toX + w * toY;

    // Mapping from square type to one of the following:
    // 0 = Square is completely inaccessible (i.e. there is no movement into or out of it which takes no more than max points).
    // 1 = Square is accessible but its position does not matter.
    // 3 = Square is accessible and its position matters.
    wxUint8 typeflags[ntypes];
    // We compare stepresults and stepresultsadj for each pair of types to determine these.
    for(type = 0; type < ntypes; ++type)
    {
        for(type2 = 0; type2 < ntypes; ++type2)
        {
            distance = turnmap[stepresults[type + ntypes * type2].turnkey];
            distance2 = turnmap[stepresultsadj[type + ntypes * type2].turnkey];
            if(distance <= max || distance2 <= max)
            {
                typeflags[type] |= 1;
                typeflags[type2] |= 1;
                if(distance != distance2)
                {
                    typeflags[type] |= 2;
                    typeflags[type2] |= 2;
                }
            }
        }
    }

    // The set of squares which are unvisited nodes.
    // To remove the node with index n, the following code is used: unvisited[n] = unvisited[--nunvisited];
    // This works, as the order of nodes in the array does not matter.
    wxUint16 unvisited[nsquares];
    wxUint16 nunvisited = 0;
    // We initialise this by checking all the board squares.
    // Squares with typeflags 0 should not be added, as they are completely inaccessible so it would be a waste of time to check them.
    // After adding a square, bit 1 is cleared in its typeflags.
    // This means that types for which position does not matter only get one square added (as their flags will change from 1 to 0),
    // while those for which position does matter will get all squares added (as their flags will change from 3 to 2).
    // We must always include the start square (not the finish, as we handle that before the main loop).
    unvisited[nunvisited] = fromSquare;
    ++nunvisited;
    typeflags[board[fromSquare]] &= ~1;
    for(square = 0; square < nsquares; ++square)
    {
        if(square == fromSquare || square == toSquare)
            continue; // The start square is added before the loop, and the finish square should not be added.
        if(typeflags[board[square]] == 0)
            continue; // Type either inaccessible, or non-position-dependent and already accounted for.
        unvisited[nunvisited] = square;
        ++nunvisited;
        typeflags[board[square]] &= ~1;
    }

    // The shortest route found from each board square to the finish square.
    // distances holds the number of die points taken to get there.
    // warptype holds the type of square that will be warped to in the final move.
    wxUint8 distances[nsquares];
    wxUint8 warptypes[nsquares];

    // Initialisation of distances and warptype.
    // This is really the first iteration of Dijkstra's algorithm, with the node considered being the finish square.
    // We allow warp moves.
    square2 = toSquare;
    for(iunvisited = 0; iunvisited < nunvisited; ++iunvisited)
    {
        square = unvisited[iunvisited];
        stepresult = getStepResult(square % w, square / w, board[square], square2 % w, square2 / w, board[square2]);
        distances[square] = turnmap[stepresult->turnkey];
        warptypes[square] = stepresult->warptype;
    }
    distances[toSquare] = 0;
    warptypes[toSquare] = 0;

    // The main loop of Dijkstra's algorithm.
    while(nunvisited > 0)
    {
        // Find the unvisited node with the shortest distance to it found so far.
        // Since no path can reduce distance, no node can possibly have a smaller distance than this one.
        // FIXME: Consider using some kind of sorted list to speed this up.
        iunvisited2 = 0;
        distance = distances[unvisited[iunvisited2]];
        for(iunvisited = 1; iunvisited < nunvisited; ++iunvisited)
        {
            if(distances[unvisited[iunvisited]] < distance)
            {
                iunvisited2 = iunvisited;
                distance = distances[unvisited[iunvisited2]];
            }
        }
        // If this square exceeds the maximum distance, give up. :(
        if(distance > max)
            break;
        // If the square with the shortest distance is the start, or the start's distance is no more than it,
        // then we have found the shortest route from the start. :)
        if(distance == distances[fromSquare])
        {
            square = randSquare(warptypes[fromSquare]);
            if(square < nsquares)
            {
                if(finalX != NULL)
                    *finalX = square % w;
                if(finalY != NULL)
                    *finalY = square / w;
            }
            return distances[fromSquare];
        }
        // Get the square number and mark the node as visited.
        square2 = unvisited[iunvisited2];
        unvisited[iunvisited2] = unvisited[--nunvisited];
        // Check each unvisited square to see if there is a better route from it to the finish via square2 than
        // the best found so far.
        for(iunvisited = 0; iunvisited < nunvisited; ++iunvisited)
        {
            square = unvisited[iunvisited];
            stepresult = getStepResult(square % w, square / w, board[square], square2 % w, square2 / w, board[square2]);
            if(stepresult->warptype < ntypes)
                continue; // Warp moves are not allowed.
            distance2 = distance + turnmap[stepresult->turnkey];
            if(distance2 < distances[square])
            {
                // Since we have found a shorter route from the square, change the square's distance to reflect this.
                distances[square] = distance2;
                // Copy the warp type (if A -> B has no warp, but B -> C leads to a certain warp, then A -> B -> C will lead to the same warp).
                warptypes[square] = warptypes[square2];
            }
        }
    }

    // No route was found.
    return (wxUint8) -1;
}

wxUint8 Board::getSuspectAt(wxUint8 x, wxUint8 y) const
{
    if(locations == NULL)
        return 255;
    wxUint16 square = y * w + x;
    for(wxUint8 n = 0; n < offsetw; ++n)
    {
        if(locations[n] == square)
            return n;
    }
    return 255;
}

bool Board::setSuspectPosition(wxUint8 n, wxUint8 x, wxUint8 y)
{
    if(x >= w || y >= h || board == NULL || n >= offsetw)
        return FALSE;
    locations[n] = y * w + x;
    return TRUE;
}

bool Board::roomWarp(wxUint8 suspect, wxUint8 room)
{
    if(suspect >= offsetw || room < offsetr || room >= ncards)
        return FALSE;
    if(getSuspectRoom(suspect) == room)
        return TRUE;
    wxUint8 matchtypes[ntypes];
    wxUint8 typecount = 0;
    for(wxUint8 type = 0; type < ntypes; ++type)
    {
        if(types[type] == room)
        {
            matchtypes[typecount] = type;
            ++typecount;
        }
    }
    wxUint16 warpdest = randSquare(matchtypes, typecount);
    if(warpdest < (w * h))
    {
        locations[suspect + offsetw] = locations[suspect] = warpdest;
        return TRUE;
    }
    return FALSE;
}

wxUint16 Board::countSquares(wxUint8 *matchtypes, wxUint8 typecount, bool needempty) const
{
    if(typecount == 0)
        return 0;
    wxUint16 count = 0;
    wxUint8 x, y, n;
    wxUint8 type;
    for(y = 0; y < h; ++y)
    {
        for(x = 0; x < w; ++x)
        {
            type = board[y * w + x];
            for(n = 0; n < typecount && type != matchtypes[n]; ++n);
            if(n < typecount)
            {
                if(needempty)
                {
                    for(n = 0; n < offsetw && locations[n] != (y * w + x); ++n);
                    if(n < offsetw)
                        continue;
                }
                ++count;
            }
        }
    }
    return count;
}

#if 0
wxUint16 Board::findSquare(wxUint8 *matchtypes, wxUint8 typecount, bool needempty) const
{
    if(typecount == 0)
        return (wxUint16) -1;
    wxUint8 x, y, n;
    wxUint8 type;
    for(y = 0; y < h; ++y)
    {
        for(x = 0; x < w; ++x)
        {
            type = board[y * w + x];
            for(n = 0; n < typecount && type != matchtypes[n]; ++n);
            if(n < typecount)
            {
                if(needempty)
                {
                    for(n = 0; n < offsetw && locations[n] != (y * w + x); ++n);
                    if(n < offsetw)
                        continue;
                }
                return y * w + x;
            }
        }
    }
    return (wxUint16) -1;
}
#endif

wxUint16 Board::randSquare(wxUint8 *matchtypes, wxUint8 typecount, bool needempty) const
{
    wxUint16 count = countSquares(matchtypes, typecount, needempty);
    if(count == 0)
        return (wxUint16) -1;
    wxUint8 x, y, n;
    wxUint8 type;
    for(y = 0; y < h; ++y)
    {
        for(x = 0; x < w; ++x)
        {
            type = board[y * w + x];
            for(n = 0; n < typecount && type != matchtypes[n]; ++n);
            if(n < typecount)
            {
                if(needempty)
                {
                    for(n = 0; n < offsetw && locations[n] != (y * w + x); ++n);
                    if(n < offsetw)
                        continue;
                }
                if(std::rand() / (RAND_MAX / count + 1) == 0)
                    return y * w + x;
                --count;
            }
        }
    }
    // It shouldn't be possible to reach this point.
    // When count reaches 1, the condition in the loop will always be true.
    return (wxUint16) -1;
}

#if 0
MoveRule *Board::getMoveRule(wxUint8 x1, wxUint8 y1, wxUint8 x2, wxUint8 y2) const
{
    if(x1 >= w || y1 >= h || x2 >= w || y2 >= h)
        return NULL;
    wxUint8 from = board[y1 * w + x1];
    wxUint8 to = board[y2 * w + x2];
    if(from > ntypes || to > ntypes)
        return NULL;
    wxUint8 *typecheck;
    wxUint8 n;
    for(n = 0; n < nmoverules; ++n)
    {
        if(moverules[n].needadj && (x1 != x2 || (y1 != y2 + 1 && y2 != y1 + 1)) && (y1 != y2 || (x1 != x2 + 1 && x2 != x1 + 1)))
            continue;
        if(moverules[n].types != NULL)
        {
            typecheck = moverules[n].types;
            if(*typecheck != 255)
            {
                // Search for either the source square type or 255.
                while(*typecheck != 255 && *typecheck != from)
                    ++typecheck;
                // Give up if 255 was reached without the source square type being found.
                if(*typecheck == 255)
                    continue;
            }
            // Skip to the 255.
            while(*typecheck != 255)
                ++typecheck;
            ++typecheck;
            if(*typecheck != 255)
            {
                // Check the destination square type in the same way as the source.
                while(*typecheck != 255 && *typecheck != to)
                    ++typecheck;
                if(*typecheck == 255)
                    continue;
            }
        }
        return moverules + n;
    }
    return NULL;
}
#endif
