// MurderSpot - a free game based on Clue/Cluedo
// Copyright 2005-2009 Christopher Elsby <chrise@chrise.me.uk>
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "global.h"

#include "boardwin.h"
#include "game.h"
#include "board.h"
#include "cmdids.h"
#include "guidefs.h"
#include <wx/gdicmn.h>
#include <wx/colour.h>
#include <wx/scrolwin.h>
#include <wx/dcclient.h>
#include <wx/bitmap.h>
#include <wx/image.h>
#include <wx/panel.h>
#include <wx/sizer.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/button.h>
#include <wx/string.h>
#include <wx/filedlg.h>
#include <wx/filename.h>
#include <wx/log.h>

BEGIN_EVENT_TABLE(BoardWin, wxScrolledWindow)
    EVT_LEFT_DOWN(BoardWin::OnMouseButton)
    EVT_BUTTON(-1, BoardWin::OnCommand)
    EVT_UPDATE_UI(-1, BoardWin::OnUpdateUI)
END_EVENT_TABLE()

BoardWin::BoardWin(wxWindow *parent, Game *gam)
    : wxScrolledWindow(parent), game(gam), bmpBoard(NULL), bmpSuspects(NULL)
{
    selectionPanel = new wxPanel(this, -1, wxPoint(CONTROL_SPACING, CONTROL_SPACING), wxSize(100, 100));
    wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
    sizer->Add(new wxStaticText(selectionPanel, -1, wxT("Choose the board image to use for the game."
        "\nYou can leave the width and height fields blank to use the defaults.")), 0, wxALL, CONTROL_SPACING);
    wxBoxSizer *sizer2 = new wxBoxSizer(wxHORIZONTAL);
    sizer2->Add(fileInput = new wxTextCtrl(selectionPanel, -1), 1, wxALL, CONTROL_SPACING);
    sizer2->Add(new wxButton(selectionPanel, ID_BROWSE, wxT("B&rowse")), 0, wxALL, CONTROL_SPACING);
    sizer->Add(sizer2, 0, wxGROW);
    wxBoxSizer *sizer3 = new wxBoxSizer(wxHORIZONTAL);
    sizer3->Add(new wxStaticText(selectionPanel, -1, wxT("Cell Dimensions (optional):")), 0, wxALL, CONTROL_SPACING);
    sizer3->Add(widthInput = new wxTextCtrl(selectionPanel, -1, wxT(""), wxDefaultPosition, wxDefaultSize, 0), 0, wxALL, CONTROL_SPACING);
    sizer3->Add(new wxStaticText(selectionPanel, -1, wxT("x")), 0, wxALL, CONTROL_SPACING);
    sizer3->Add(heightInput = new wxTextCtrl(selectionPanel, -1, wxT(""), wxDefaultPosition, wxDefaultSize, 0), 0, wxALL, CONTROL_SPACING);
    sizer->Add(sizer3, 0, wxGROW);
    wxBoxSizer *sizer4 = new wxBoxSizer(wxHORIZONTAL);
    sizer4->Add(new wxButton(selectionPanel, wxID_OK), 0, wxALL, CONTROL_SPACING);
    sizer4->Add(new wxButton(selectionPanel, wxID_CANCEL), 0, wxALL, CONTROL_SPACING);
    sizer->Add(sizer4, 0, wxGROW);
    selectionPanel->SetSizerAndFit(sizer);
    updateImage();
}

BoardWin::~BoardWin()
{
    if(bmpBoard != NULL)
        delete bmpBoard;
    if(bmpSuspects != NULL)
        delete[] bmpSuspects;
}

void BoardWin::setGame(Game *gam)
{
    game = gam;
}

void BoardWin::updateScrollBars()
{
    int w = 0;
    int h = 0;
    if(game != NULL && bmpBoard != NULL)
    {
        w = game->getBoard()->imageWidth;
        if(w < 1)
            w = 1;
        h = game->getBoard()->imageHeight;
        if(h < 1)
            h = 1;
        SetScrollRate(w, h);
        w *= game->getBoard()->getWidth();
        h *= game->getBoard()->getHeight();
    }
    else
    {
        SetScrollRate(CONTROL_SPACING, CONTROL_SPACING);
    }
    if(selectionPanel->IsShown())
    {
        int w2 = selectionPanel->GetRect().GetRight();
        if(w2 > w)
            w = w2;
        int h2 = selectionPanel->GetRect().GetBottom();
        if(h2 > h)
            h = h2;
    }
    SetVirtualSize(w, h);
}

//bool BoardWin::setImage(const wxImage *img, wxUint8 tilew, wxUint8 tileh)
bool BoardWin::updateImage()
{
    if(game == NULL || game->getBoard()->imageFile.empty())
    {
        // Delete bitmaps and display the selection panel.
        if(bmpBoard != NULL)
            delete bmpBoard;
        bmpBoard = NULL;
        if(bmpSuspects != NULL)
            delete[] bmpSuspects;
        bmpSuspects = NULL;
        changeImage();
        return TRUE;
    }
    // Load the image file.
    wxImage img(game->getBoard()->imageFile);
    if(!img.Ok())
        return FALSE;
    // Fail if the needed board parameters are not available.
    if(game == NULL || game->getBoard()->getWidth() == 0 || game->getBoard()->getHeight() == 0 || game->getBoard()->suspectCount() == 0)
        return FALSE;
    // Fail if the image is not big enough to hold the board with the suspects below it, with at least 1 pixel per square.
    if(img.GetWidth() < game->getBoard()->getWidth() || img.GetWidth() < game->getBoard()->suspectCount() * 2 + 1 || img.GetHeight() < game->getBoard()->getHeight() + 1)
        return FALSE;
    // Now there are no more opportunities for failure, delete the old bitmaps.
    if(bmpBoard != NULL)
        delete bmpBoard;
    bmpBoard = NULL;
    if(bmpSuspects != NULL)
        delete[] bmpSuspects;
    bmpSuspects = NULL;
    // Determine the square size (it doesn't actually have to be square :-P).
    // Note that the lack of +1 here is intentional --------------------------------------------------------------------------------------------|
    // The extra pixel needed for the background contributes to the minimum allowed width,                                                      |
    // but it is not actually part of the grid, so should not be considered when sizing it.                                                     V
    wxUint8 w = img.GetWidth() / (game->getBoard()->suspectCount() * 2 + 1 > game->getBoard()->getWidth() ? game->getBoard()->suspectCount() * 2 : game->getBoard()->getWidth());
    if(game->getBoard()->imageWidth == 0)
        game->getBoard()->imageWidth = w;
    wxUint8 h = img.GetHeight() / (game->getBoard()->getHeight() + 1);
    if(game->getBoard()->imageHeight == 0)
        game->getBoard()->imageHeight = h;
    // Grab the board image.
    bmpBoard = new wxBitmap(img.GetSubImage(wxRect(0, 0, w * game->getBoard()->getWidth(), h * game->getBoard()->getHeight())).Scale(game->getBoard()->imageWidth * game->getBoard()->getWidth(),
                game->getBoard()->imageHeight * game->getBoard()->getHeight()));
    // Grab the suspect images.
    bmpSuspects = new wxBitmap[game->getBoard()->suspectCount() * 2];
    for(wxUint8 suspect = 0; suspect < game->getBoard()->suspectCount(); ++suspect)
    {
        bmpSuspects[suspect] = wxBitmap(img.GetSubImage(wxRect(w * suspect, h * game->getBoard()->getHeight(), w, h)).Scale(game->getBoard()->imageWidth, game->getBoard()->imageHeight));
        bmpSuspects[game->getBoard()->suspectCount() + suspect] = wxBitmap(img.GetSubImage(wxRect(w * (game->getBoard()->suspectCount() + suspect),
                    h * game->getBoard()->getHeight(), w, h)).Scale(game->getBoard()->imageWidth, game->getBoard()->imageHeight));
        bmpSuspects[suspect].SetMask(new wxMask(bmpSuspects[game->getBoard()->suspectCount() + suspect], *wxBLACK));
        bmpSuspects[game->getBoard()->suspectCount() + suspect].SetMask(new wxMask(bmpSuspects[game->getBoard()->suspectCount() + suspect], *wxBLACK));
    }
#if 0 // Causes trouble with the selection panel on Windows.
    // Take the bottom-right pixel (of the original image, not the board) as the background color.
    int x = img.GetWidth() - 1;
    int y = img.GetHeight() - 1;
    SetBackgroundColour(wxColour(img.GetRed(x, y), img.GetGreen(x, y), img.GetBlue(x, y)));
#endif
    // Set scroll bars.
    updateScrollBars();
    // Close the selection panel.
    closeChangeImage();
    // Return success.
    return TRUE;
}

void BoardWin::changeImage()
{
    if(game != NULL)
    {
        fileInput->SetValue(game->getBoard()->imageFile);
        wxString str;
        if(game->getBoard()->imageWidth)
            str.Printf(wxT("%i"), game->getBoard()->imageWidth);
        widthInput->SetValue(str);
        if(game->getBoard()->imageHeight)
            str.Printf(wxT("%i"), game->getBoard()->imageHeight);
        heightInput->SetValue(str);
    }
    selectionPanel->Show();
    updateScrollBars();
    Scroll(0, 0);
}

void BoardWin::closeChangeImage()
{
    selectionPanel->Hide();
    updateScrollBars();
    Refresh();
}

void BoardWin::OnDraw(wxDC &dc)
{
    if(bmpBoard != NULL)
        dc.DrawBitmap(*bmpBoard, 0, 0, FALSE);
    if(bmpSuspects != NULL && game != NULL && game->getBoard()->hasLocations() && game->getBoard()->suspectCount() > 0)
    {
        wxUint8 w = bmpSuspects->GetWidth();
        wxUint8 h = bmpSuspects->GetHeight();
        for(wxUint8 suspect = game->getBoard()->suspectMin(); suspect < game->getBoard()->suspectMax(); ++suspect)
        {
            dc.DrawBitmap(bmpSuspects[suspect - game->getBoard()->suspectMin()], game->getBoard()->getSuspectX(suspect) * w, game->getBoard()->getSuspectY(suspect) * h, TRUE);
#if 0
            if(game->getBoard()->isMoving(suspect))
                dc.DrawBitmap(bmpSuspects[game->getBoard()->suspectCount() + suspect - game->getBoard()->suspectMin()], game->getBoard()->getOriginalX(suspect) * w, game->getBoard()->getOriginalY(suspect) * h, TRUE);
#endif
        }
    }
}

void BoardWin::OnMouseButton(wxMouseEvent &event)
{
    if(!selectionPanel->IsShown() && event.LeftDown() && game != NULL && game->getBoard()->imageWidth > 0 && game->getBoard()->imageHeight > 0)
    {
        int x, y;
        GetViewStart(&x, &y);
        x += event.GetX() / game->getBoard()->imageWidth;
        y += event.GetY() / game->getBoard()->imageHeight;
        if(game != NULL && game->canControl() && game->move(x, y))
            game->update();
    }
    else
    {
        event.Skip();
    }
}

void BoardWin::OnCommand(wxCommandEvent &event)
{
    wxFileName old;
    wxString filename;
    switch(event.GetId())
    {
    case wxID_OK:
        if(game != NULL)
        {
            filename = fileInput->GetValue();
            if(filename.empty())
            {
                wxLogError(wxT("No filename specified."));
            }
            else
            {
                unsigned long w = 0, h = 0;
                widthInput->GetValue().ToULong(&w);
                heightInput->GetValue().ToULong(&h);
                game->getBoard()->setImageFile(filename, w, h);
                if(!updateImage())
                    wxLogError(wxT("Failed to load image."));
            }
        }
        break;
    case wxID_CANCEL:
        if(bmpBoard != NULL)
            closeChangeImage();
        break;
    case ID_BROWSE:
        old = fileInput->GetValue();
        filename = wxFileSelector(wxT("Choose Board Image"), old.GetPath(wxPATH_GET_VOLUME), old.GetFullName(), wxT(""),
                wxT("All Files (*)|*"), wxOPEN | wxFILE_MUST_EXIST, this);
        if(!filename.empty() && filename != fileInput->GetValue())
        {
            fileInput->SetValue(filename);
            widthInput->SetValue(wxT(""));
            heightInput->SetValue(wxT(""));
        }
        break;
    default:
        event.Skip();
    }
}

void BoardWin::OnUpdateUI(wxUpdateUIEvent &event)
{
    switch(event.GetId())
    {
    case wxID_OK:
        event.Enable(!fileInput->GetValue().empty());
        break;
    case wxID_CANCEL:
        event.Enable(bmpBoard != NULL);
        break;
    case ID_BROWSE:
        break;
    default:
        event.Skip();
    }
}
