// MurderSpot - a free game based on Clue/Cluedo
// Copyright 2005-2009 Christopher Elsby <chrise@chrise.me.uk>
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "global.h"

#include "chatwin.h"
#include "guidefs.h"
#include "mainwin.h"
#include "network.h"
#include <wx/sizer.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/colour.h>

static wxString nextParam(wxString &msg)
{
    while(!msg.empty() && msg.at(0) == ' ')
        msg.erase(0, 1);
    if(msg.empty())
        return wxEmptyString;
    wxString param;
    size_t pos = msg.find(wxT(" "));
    if(pos < wxString::npos)
    {
        param = msg.substr(0, pos);
        msg.erase(0, pos + 1);
    }
    else
    {
        param = msg;
        msg.clear();
    }
    return param;
}

BEGIN_EVENT_TABLE(ChatWin, wxPanel)
    EVT_TEXT_ENTER(-1, ChatWin::OnCommand)
    EVT_SET_FOCUS(ChatWin::OnFocus)
END_EVENT_TABLE()

ChatWin::ChatWin(wxWindow *parent, MainWin *mainwin)
    : wxPanel(parent), main(mainwin)
{
    wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
    sizer->Add(history = new wxTextCtrl(this, -1, wxT(""), wxDefaultPosition, wxDefaultSize, wxTE_MULTILINE | wxTE_READONLY), 1, wxGROW);
#if 0 // Doesn't work very well
    history->SetBackgroundColour(*wxBLACK);
    history->SetForegroundColour(*wxWHITE);
    history->SetDefaultStyle(wxTextAttr(*wxWHITE, *wxBLACK));
#endif
    wxBoxSizer *inputsizer = new wxBoxSizer(wxHORIZONTAL);
    wxString nick;
    if(main != NULL)
    {
        Network *net = main->getNetwork();
        if(net != NULL)
            nick = net->getNick();
    }
    inputsizer->Add(nickdisplay = new wxStaticText(this, -1, nick), 0, wxALL, 5);
    inputsizer->Add(input = new wxTextCtrl(this, -1, wxT(""), wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER | wxTE_NOHIDESEL), 1);
    sizer->Add(inputsizer, 0, wxALL | wxGROW, CONTROL_SPACING);
    SetSizerAndFit(sizer);
}

ChatWin::~ChatWin()
{
    if(main != NULL && !sendto.empty() && sendto.at(0) == '#')
    {
        Network *net = main->getNetwork();
        if(net != NULL)
            net->command(wxT("PART ") + sendto + wxT(" Leaving"));
    }
}

void ChatWin::changeNick(wxString newnick)
{
    nickdisplay->SetLabel(newnick);
    Layout();
}

void ChatWin::receiveChat(wxString sender, wxString target, wxString type, wxString msg)
{
    wxString nick;
    if(main != NULL)
    {
        Network *net = main->getNetwork();
        if(net != NULL)
            nick = net->getNick();
    }
    type.MakeUpper();
    wxString display;
    if(!sendto.empty() && !target.empty() && target.Lower() != ((sender.Lower() == sendto.Lower()) ? nick : sendto).Lower())
        display = wxT(">") + target + wxT("< ");
    if(type == wxT("STATUS") || type == wxT("LISTSTART") || type == wxT("LIST") || type == wxT("LISTEND"))
        display += wxT("*** ") + msg;
    else if(type == wxT("NAMES"))
        display = wxT("*** Users on ") + target + wxT(": ") + msg;
    else if(type == wxT("MODE"))
    {
        if(sender.empty())
            display = wxT("*** Modes on ") + target + wxT(": ") + msg;
        else
            display = wxT("*** Mode change on ") + target + wxT(" by ") + sender + wxT(": ") + msg;
    }
    else if(type == wxT("TOPIC"))
    {
        if(msg.empty())
            display = sender.empty() ? (wxT("*** Topic for ") + target + wxT(" is unset"))
                                     : ((wxT("*** ") + sender + wxT(" removed the topic for ")) + target);
        else
            display = (sender.empty() ? wxT("*** Topic for ") : (wxT("*** ") + sender + wxT(" set the topic for "))) + target + wxT(": ") + msg;
    }
    else if(type == wxT("PRIVMSG"))
        display += wxT("<") + sender + wxT("> ") + msg;
    else if(type == wxT("ACTION"))
        display += wxT("* ") + sender + wxT(" ") + msg;
    else if(type == wxT("NOTICE"))
        display += wxT("[") + sender + wxT("] ") + msg;
    else if(type == wxT("CTCP"))
        display += wxT("<") + sender + wxT(" CTCP> ") + msg;
    else if(type == wxT("NCTCP"))
        display += wxT("[") + sender + wxT(" CTCP] ") + msg;
    else if(type == wxT("JOIN"))
        display = wxT("--> ") + sender + wxT(" has joined ") + target + (msg.empty() ? wxT("") : wxT(" (") + msg + wxT(")"));
    else if(type == wxT("PART"))
        display = wxT("<-- ") + sender + wxT(" has left ") + target + (msg.empty() ? wxT("") : wxT(" (") + msg + wxT(")"));
    else if(type == wxT("KICK"))
    {
        wxString affected = nextParam(msg);
        display = wxT("<-- ") + affected + wxT(" has been kicked from ") + target + wxT(" by ") + sender +
            (msg.empty() ? wxT("") : wxT(" (") + msg + wxT(")"));
    }
    else if(type == wxT("QUIT"))
        display = wxT("<-- ") + sender + wxT(" has quit ") + (msg.empty() ? wxT("") : wxT(" (") + msg + wxT(")"));
    else if(type == wxT("NICK"))
    {
        if(sender == nick)
            display = wxT("--- You have changed nick to ") + msg;
        else
            display = wxT("--- ") + sender + wxT(" has changed nick to ") + msg;
    }
    else if(type == wxT("RAW"))
        display = msg;
    else
        return;
    history->AppendText(display + wxT("\n"));
    // Workaround for a graphical glitch under Windows.
    wxWindow *parent = GetParent();
    if(parent == NULL)
        Refresh();
    else
        parent->Refresh();
}

void ChatWin::OnCommand(wxCommandEvent &event)
{
    if(event.GetEventObject() == input && event.GetEventType() == wxEVT_COMMAND_TEXT_ENTER && main != NULL)
    {
        wxString request = input->GetValue();
        if(!request.empty())
        {
            input->Clear();
            if(request.Upper() == wxT("/CLEAR"))
            {
                history->Clear();
            }
            else
            {
                Network *net = main->getNetwork();
                if(net != NULL)
                {
                    if(request.at(0) == wxT('/'))
                    {
                        request.erase(0, 1);
                        if(!request.empty() && request.at(0) == wxT('/'))
                        {
                            receiveChat(net->getNick(), sendto, wxT("PRIVMSG"), request);
                            net->chat(sendto, request);
                        }
                        else
                        {
                            wxString type = nextParam(request);
                            type.MakeUpper();
                            if(type == wxT("J"))
                                type = wxT("JOIN");
                            if(type == wxT("QUOTE"))
                            {
                                net->command(request);
                            }
                            else if(type == wxT("RAW"))
                            {
                                net->immediatecommand(request);
                            }
                            else if(type == wxT("ME") || type == wxT("ACTION"))
                            {
                                receiveChat(net->getNick(), sendto, wxT("ACTION"), request);
                                net->chat(sendto, wxT("\x1") wxT("ACTION ") + request + wxT("\x1"));
                            }
                            else if(type == wxT("CTCP"))
                            {
                                wxString target = nextParam(request);
                                if(!target.empty())
                                {
                                    receiveChat(net->getNick(), target, wxT("CTCP"), request);
                                    net->chat(target, wxT("\x1") + request + wxT("\x1"));
                                }
                            }
                            else if(type == wxT("NCTCP"))
                            {
                                wxString target = nextParam(request);
                                if(!target.empty())
                                {
                                    receiveChat(net->getNick(), target, wxT("NCTCP"), request);
                                    net->command(wxT("NOTICE ") + target + wxT(" :\x1") + request + wxT("\x1"));
                                }
                            }
                            else if(type == wxT("MSG") || type == wxT("PRIVMSG"))
                            {
                                wxString target = nextParam(request);
                                if(!target.empty())
                                {
                                    receiveChat(net->getNick(), target, wxT("PRIVMSG"), request);
                                    net->chat(target, request);
                                }
                            }
                            else if(type == wxT("NOTICE"))
                            {
                                wxString target = nextParam(request);
                                if(!target.empty())
                                {
                                    receiveChat(net->getNick(), target, wxT("NOTICE"), request);
                                    net->command(wxT("NOTICE ") + target + wxT(" :") + request);
                                }
                            }
                            else if(type == wxT("QUERY") || type == wxT("DIALOG"))
                            {
                                wxString target = nextParam(request);
                                if(!target.empty())
                                    main->receiveChat(net, target, net->getNick(), wxT("QUERY"), wxT(""));
                            }
                            else if(type == wxT("NAMES") || type == wxT("JOIN"))
                            {
                                wxString target = nextParam(request);
                                net->command(type + wxT(" ") + (target.empty() ? sendto : target));
                            }
                            else if(type == wxT("TOPIC"))
                            {
                                wxString target;
                                if(request.at(0) == wxT('#'))
                                    target = nextParam(request);
                                else
                                    target = sendto;
                                net->command(type + wxT(" ") + target + (request.empty() ? wxT("") : (wxT(" :") + request)));
                            }
                            else if(type == wxT("PART"))
                            {
                                wxString target;
                                if(request.at(0) == wxT('#'))
                                    target = nextParam(request);
                                else
                                    target = sendto;
                                net->command(type + wxT(" ") + target + (request.empty() ? wxT(" :Leaving") : (wxT(" :") + request)));
                            }
                            else if(type == wxT("CYCLE"))
                            {
                                wxString target;
                                if(request.at(0) == wxT('#'))
                                    target = nextParam(request);
                                else
                                    target = sendto;
                                net->command(wxT("PART ") + target + (request.empty() ? wxT(" :Rejoining") : (wxT(" :") + request)));
                                net->command(wxT("JOIN ") + target);
                            }
                            else
                            {
                                net->command(type + wxT(" ") + request);
                            }
                        }
                    }
                    else
                    {
                        receiveChat(net->getNick(), sendto, wxT("PRIVMSG"), request);
                        net->chat(sendto, request);
                    }
                }
            }
        }
    }
    else
    {
        event.Skip();
    }
}

void ChatWin::OnFocus(wxFocusEvent &event)
{
    input->SetFocus();
}

