// MurderSpot - a free game based on Clue/Cluedo
// Copyright 2005-2009 Christopher Elsby <chrise@chrise.me.uk>
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "global.h"

#include "game.h"
#include "board.h"
#include "manager.h"
#include "network.h"
#include <cstdlib>
#include <wx/timer.h>
#include <wx/filename.h>

static wxString nextParam(wxString &msg)
{
    while(!msg.empty() && msg.at(0) == wxT(' '))
        msg.erase(0, 1);
    if(msg.empty())
        return wxEmptyString;
    wxString param;
    size_t pos = msg.find(wxT(" "));
    if(pos < wxString::npos)
    {
        param = msg.substr(0, pos);
        msg.erase(0, pos + 1);
    }
    else
    {
        param = msg;
        msg.clear();
    }
    return param;
}

static wxString format_turn_config(wxString name, wxUint8 points)
{
    if(points == 0)
        name << wxT(" does not require any points.");
    else if(points == 1)
        name << wxT(" requires 1 point.");
    else if(points < 255)
        name << wxT(" requires ") << (int) points << wxT(" points.");
    else
        name << wxT(" is disabled.");
    return name;
}

class GameTimer : public wxTimer
{
public:
    GameTimer(Game *gam) : game(gam) { }
    virtual void Notify();
    bool Start(bool reset = TRUE)
        {return game == NULL || !game->getControlling() || game->isPaused() || (!reset && IsRunning()) || wxTimer::Start(5000, TRUE);}
    Game *game;
};

Game::Game()
    : roomWarp(TRUE), autoClearLog(FALSE), startSuggest(FALSE), startSuggestWarped(TRUE), mgr(NULL), paused(FALSE), gamestate(0), sSuspect(255)
    , sWeapon(255), sRoom(255) , sDisprove(255), disprover(255), candisprove(FALSE), players(NULL), cards(NULL), active(255), diceroll(0), dicepoints(0)
    , changedRooms(FALSE), channel(NULL), controlling(TRUE), controller(NULL), timer(NULL)
{
    timer = new GameTimer(this);
}

Game::~Game()
{
    if(timer != NULL)
        delete timer;
#if 0
    if(inProgress() && channel != NULL)
        channel->sendgamerequest(getGameID(), wxT("ENDGAME"));
#endif
    if(gamestate != 3)
    {
        gamestate = 3;
        updateTableTopic();
    }
    if(players != NULL)
        delete[] players;
    if(cards != NULL)
        delete[] cards;
}

void Game::pauseGame(bool pause)
{
    if(pause == paused)
        return;
    paused = pause;
    if(gamestate != 1)
        return;
    if(pause)
    {
        if(timer != NULL)
            timer->Stop();
        update();
        logNotice(wxT("Game paused."), TRUE, FALSE);
    }
    else
    {
        if(timer != NULL)
            timer->Start();
        update();
        logNotice(wxT("Game resumed."), TRUE, FALSE);
        postStatus();
    }
    if(controlling && channel != NULL)
        channel->sendgamerequest(getGameID(), pause ? wxT("PAUSE") : wxT("UNPAUSE"));
    updateTableTopic();
}

void Game::stopGame(bool updatestate)
{
    if(timer != NULL)
        timer->Stop();
    if(gamestate == 1 && channel != NULL)
    {
        if(controlling)
            channel->sendgamerequest(getGameID(), wxT("ENDGAME"));
        else if(controller != NULL)
            controller->sendgamerequest(getGameID(), wxT("LEAVEGAME"));
    }
    active = disprover = 255;
    sSuspect = sWeapon = sRoom = sDisprove = 255;
    diceroll = dicepoints = 0;
    changedRooms = FALSE;
    if(updatestate && gamestate != 3)
    {
        gamestate = 3;
        updateTableTopic();
    }
}

void Game::reset()
{
    stopGame(FALSE);
    if(players != NULL)
        delete[] players;
    if(cards != NULL)
        delete[] cards;
    if(board.suspectCount() > 0)
    {
        players = new Player[board.suspectCount()];
        cards = new wxUint8[board.cardCount()];
        for(int i = 0; i < board.cardCount(); ++i)
            cards[i] = 255;
    }
    else
    {
        players = NULL;
        cards = NULL;
    }
    clearLog();
    if(gamestate != 0)
    {
        gamestate = 0;
        updateTableTopic();
    }
}

wxString Game::getGameID() const
{
    if(channel == NULL)
        return wxT("local");
    return channel->name;
}

Network *Game::getNetwork() const
{
    if(channel == NULL)
        return NULL;
    return channel->getNetwork();
}

void Game::processNet(const NetUser *sender, wxString type, wxString msg)
{
    if(sender == NULL || sender->getNetwork() == NULL || channel == NULL || channel->getNetwork() != sender->getNetwork())
        return;
    if(controlling)
    {
        if(type == wxT("ENDGAME") || type == wxT("LEAVEGAME"))
        {
            // A player has left the game.
            if(players != NULL)
            {
                bool lostplayer = FALSE;
                for(int i = board.suspectMin(); i < board.suspectMax(); ++i)
                {
                    if(players[i].netuser == sender)
                    {
                        if(players[i].playing == 1)
                            lostplayer = TRUE;
                        players[i].netuser = NULL;
                        players[i].type = playerTypeFallback;
                        if(started())
                        {
                            wxString request = wxT("SETPLAYER ");
                            request << (int) i << wxT(" ") << (int) playerTypeFallback;
                            request << wxT(" ") << channel->getUser()->nick << wxT(" ") << players[i].name;
                            channel->sendgamerequest(getGameID(), request);
                        }
                        if(inProgress() && i == getCurrentPlayer() && players[i].isAI() && timer != NULL)
                            timer->Start();
                    }
                }
                if(lostplayer && mgr != NULL)
                    mgr->playerLost(this);
            }
        }
        else if(inProgress())
        {
            // Receiving request from player.
            Player *plactive = getPlayer(active);
            Player *pldisprove = getPlayer(disprover);
            if(type == wxT("JOIN"))
            {
                // JOIN
                // Requests that the server initialise the player's game state.
                netStartGame(sender);
            }
            if(plactive != NULL && sender == plactive->netuser)
            {
                if(type == wxT("MOVE"))
                {
                    // MOVE <x> <y>
                    // Moves the current player
                    unsigned long x = 256, y = 256;
                    if(nextParam(msg).ToULong(&x) && x < 256 && nextParam(msg).ToULong(&y) && y < 256 && move(x, y))
                    {
                        update();
                        sender->respondgamerequest(getGameID(), wxT("MOVE DONE"));
                    }
                    else
                    {
                        sender->respondgamerequest(getGameID(), wxT("MOVE FAIL"));
                    }
                }
                else if(type == wxT("CANCELMOVE"))
                {
                    // CANCELMOVE
                    // Cancel's the current player's move
                    if(cancelMove())
                    {
                        update();
                        sender->respondgamerequest(getGameID(), wxT("CANCELMOVE DONE"));
                    }
                    else
                    {
                        sender->respondgamerequest(getGameID(), wxT("CANCELMOVE FAIL"));
                    }
                }
                else if(type == wxT("SUGGEST"))
                {
                    // SUGGEST <suspect> <weapon>
                    // Makes a suggestion
                    unsigned long s = 256, w = 256;
                    if(nextParam(msg).ToULong(&s) && s < 256 && nextParam(msg).ToULong(&w) && w < 256 && suggest(s, w))
                    {
                        update();
                        sender->respondgamerequest(getGameID(), wxT("SUGGEST DONE"));
                    }
                    else
                    {
                        sender->respondgamerequest(getGameID(), wxT("SUGGEST FAIL"));
                    }
                }
                else if(type == wxT("ACCUSE"))
                {
                    // ACCUSE <suspect> <weapon> <room>
                    // Makes an accusation
                    unsigned long s = 256, w = 256, r = 256;
                    if(nextParam(msg).ToULong(&s) && s < 256 && nextParam(msg).ToULong(&w) && w < 256 &&
                            nextParam(msg).ToULong(&r) && r < 256 && accuse(s, w, r))
                    {
                        update();
                        sender->respondgamerequest(getGameID(), wxT("ACCUSE DONE"));
                    }
                    else
                    {
                        sender->respondgamerequest(getGameID(), wxT("ACCUSE FAIL"));
                    }
                }
                else if(type == wxT("ENDTURN"))
                {
                    // ENDTURN
                    // Ends the current player's turn
                    if(!activeSuggestion() && startTurn())
                    {
                        update();
                        sender->respondgamerequest(getGameID(), wxT("ENDTURN DONE"));
                    }
                    else
                    {
                        sender->respondgamerequest(getGameID(), wxT("ENDTURN FAIL"));
                    }
                }
            }
            if(pldisprove != NULL && sender == pldisprove->netuser)
            {
                if(type == wxT("DISPROVE"))
                {
                    // DISPROVE <card>
                    // Disproves the current suggestion
                    unsigned long c = 256;
                    if(nextParam(msg).ToULong(&c) && c < 256 && disprove(c))
                    {
                        update();
                        sender->respondgamerequest(getGameID(), wxT("DISPROVE DONE"));
                    }
                    else
                    {
                        sender->respondgamerequest(getGameID(), wxT("DISPROVE FAIL"));
                    }
                }
            }
        }
    }
    else
    {
        // If the sender is not the controller, make them the controller if we do not yet have one, otherwise discard the message.
        if(sender != controller)
        {
            if(controller == NULL)
                controller = sender;
            else
                return;
        }
        // Receiving instructions from controller.
        if(type == wxT("BEGINGAME"))
        {
            // BEGINGAME <protocolversion> <width> <height> <suspectname> <suspectcount> <weaponname> <weaponcount> <roomname> <roomcount>
            //      <suggestturns> <accuseturns> <imagewidth> <imageheight> <imagefile>
            // Initialises board dimensions and card counts.
            unsigned long pver = 256, w = 256, h = 256, scount = 256, wcount = 256, rcount = 256, sturns = 256, aturns = 256, imgw = 0, imgh = 0;
            nextParam(msg).ToULong(&pver);
            nextParam(msg).ToULong(&w);
            nextParam(msg).ToULong(&h);
            wxString sname = nextParam(msg);
            nextParam(msg).ToULong(&scount);
            wxString wname = nextParam(msg);
            nextParam(msg).ToULong(&wcount);
            wxString rname = nextParam(msg);
            nextParam(msg).ToULong(&rcount);
            nextParam(msg).ToULong(&sturns);
            nextParam(msg).ToULong(&aturns);
            nextParam(msg).ToULong(&imgw);
            nextParam(msg).ToULong(&imgh);
            wxString imgfile = msg;
            if(pver == 1 && w < 256 && h < 256 && scount < 256 && wcount < 256 && rcount < 256 && sturns < 256 && aturns < 256
                    && imgw < 256 && imgh < 256
                    && board.init(w, h, scount, wcount, rcount, sturns, aturns)
                    && board.setNameSuspect(sname) && board.setNameWeapon(wname) && board.setNameRoom(rname))
            {
                board.setImageFile(imgfile, imgw, imgh);
                reset();
                gamestate = 1;
                if(mgr != NULL)
                    mgr->updateImage(this);
                sender->respondgamerequest(getGameID(), wxT("BEGINGAME DONE"));
            }
            else
            {
                sender->respondgamerequest(getGameID(), wxT("BEGINGAME FAIL"));
            }
        }
        else if(type == wxT("ENDGAME"))
        {
            // ENDGAME
            // Aborts the game.
            if(gamestate < 2)
            {
                stopGame();
                update();
                postStatus();
            }
        }
        else if(type == wxT("PAUSE"))
        {
            // PAUSE
            // Pauses the game.
            pauseGame();
        }
        else if(type == wxT("UNPAUSE"))
        {
            // UNPAUSE
            // Resumes the game.
            resumeGame();
        }
        else if(type == wxT("CARDNAMES"))
        {
            // CARDNAMES <n> <card n> <card n + 1> ...
            // Initialises card names.
            unsigned long n = 256;
            wxString name;
            if(nextParam(msg).ToULong(&n))
            {
                while(1)
                {
                    if(msg.empty())
                    {
                        sender->respondgamerequest(getGameID(), wxT("CARDNAMES DONE"));
                        break;
                    }
                    name = nextParam(msg);
                    name.Replace(wxT("_"), wxT(" "));
                    if(n >= 256 || !board.setCardName(n, name))
                    {
                        sender->respondgamerequest(getGameID(), wxT("CARDNAMES FAIL"));
                        break;
                    }
                    ++n;
                }
            }
            else
            {
                sender->respondgamerequest(getGameID(), wxT("CARDNAMES FAIL"));
            }
        }
        else if(type == wxT("SUSPECTPOS"))
        {
            // SUSPECTPOS <n> <suspect n x> <suspect n y> <suspect n + 1 x> <suspect n + 1 y> ...
            // Initialises suspect positions.
            unsigned long n = 256, x = 256, y = 256;
            if(nextParam(msg).ToULong(&n))
            {
                while(1)
                {
                    if(msg.empty())
                    {
                        sender->respondgamerequest(getGameID(), wxT("SUSPECTPOS DONE"));
                        break;
                    }
                    if(!nextParam(msg).ToULong(&x) || x >= 256 || !nextParam(msg).ToULong(&y) || y >= 256 ||
                            !board.setSuspectPosition(n, x, y) || !board.commitMove(n))
                    {
                        sender->respondgamerequest(getGameID(), wxT("SUSPECTPOS FAIL"));
                        break;
                    }
                    ++n;
                }
            }
            else
            {
                sender->respondgamerequest(getGameID(), wxT("SUSPECTPOS FAIL"));
            }
        }
        else if(type == wxT("SETPLAYER"))
        {
            // SETPLAYER <suspect> <type> <nick> <name>
            // Sets a player entry.
            unsigned long n = 256, type = 256;
            if(players != NULL && nextParam(msg).ToULong(&n) && n < board.suspectMax() && nextParam(msg).ToULong(&type) && type < playerTypeCount)
            {
                players[n].type = type;
                wxString netuser = nextParam(msg);
                players[n].name = msg; // May contain spaces.
                // Check whether the network user is really just us.
                const NetUser *usr = channel->getUser();
                if(usr != NULL && usr->nick == netuser)
                    players[n].netuser = NULL;
                else
                    players[n].netuser = channel->getUser(netuser);
                if(players[n].type > 0 && players[n].type < playerTypeCount)
                {
                    wxString logmsg = board.getCardName(n) + wxT(" is played by ");
                    if(players[n].type > 1)
                        logmsg += playerTypeNames[players[n].type];
                    else if(players[n].name.empty())
                    {
                        if(players[n].netuser == NULL)
                            logmsg += wxT("Local player");
                        else
                            logmsg += wxT("Network player ") + players[n].netuser->nick;
                    }
                    else
                    {
                        logmsg += players[n].name;
                        if(players[n].netuser == NULL)
                            logmsg += wxT(" (Local player)");
                        else
                            logmsg += wxT(" (Network player ") + players[n].netuser->nick + wxT(")");
                    }
                    logNotice(logmsg);
                }
            }
            else
            {
                sender->respondgamerequest(getGameID(), wxT("SETPLAYER FAIL"));
            }
        }
        else if(type == wxT("HAVECARDS"))
        {
            // HAVECARDS <suspect> <card ID...>
            // Indicates that the suspect has the specified cards.
            unsigned long n = 256, card = 256;
            if(cards != NULL && nextParam(msg).ToULong(&n) && n <= board.suspectMax())
            {
                while(1)
                {
                    if(msg.empty())
                    {
                        sender->respondgamerequest(getGameID(), wxT("HAVECARDS DONE"));
                        break;
                    }
                    if(!nextParam(msg).ToULong(&card) || card >= board.cardCount())
                    {
                        sender->respondgamerequest(getGameID(), wxT("HAVECARDS FAIL"));
                        break;
                    }
                    cards[card] = n;
                }
            }
            else
            {
                sender->respondgamerequest(getGameID(), wxT("HAVECARDS FAIL"));
            }
        }
        else if(type == wxT("SETTINGS"))
        {
            // SETTINGS <name0> <val0> <name1> <val1> ... END
            // Indicates various game options.
            // Recognised: HOSTVER ROOMWARP AUTOCLEARLOG BOARDTITLE DICE STARTSUGGEST STARTSUGGESTWARPED
            // Reserved: POINTCONFCOUNT
            wxString name, value;
            while(1)
            {
                name = nextParam(msg);
                if(name == wxT("END"))
                    break;
                value = nextParam(msg);
                if(name.empty() || value.empty())
                    break;
                name.MakeUpper();
                value.Replace(wxT("_"), wxT(" "));
                if(name == wxT("HOSTVER"))
                    controllerver = value;
                else if(name == wxT("BOARDTITLE"))
                    board.setTitle(value);
                else if(name == wxT("DICE"))
                    board.setDiceString(value);
                else if(value == wxT("1") || value == wxT("0"))
                {
                    if(name == wxT("ROOMWARP"))
                        roomWarp = (value == wxT("1"));
                    else if(name == wxT("AUTOCLEARLOG"))
                        autoClearLog = (value == wxT("1"));
                    else if(name == wxT("STARTSUGGEST"))
                        startSuggest = (value == wxT("1"));
                    else if(name == wxT("STARTSUGGESTWARPED"))
                        startSuggestWarped = (value == wxT("1"));
                }
            }
        }
        else if(type == wxT("LOGSETTINGS"))
        {
            // LOGSETTINGS
            // Indicates that the game details should be shown in the log.
            logNotice(getGameDetails());
        }
        else if(type == wxT("POINTCONF"))
        {
            // POINTCONF <n> <name n> <pointcount n> <name n+1> <pointcount n+1> ...
            // Indicates a point configuration entry.
            nextParam(msg);
            wxString name, strpointcount;
            long pointcount;
            while(1)
            {
                name = nextParam(msg);
                name.Replace(wxT("_"), wxT(" "));
                strpointcount = nextParam(msg);
                if(name.empty() || strpointcount.empty())
                    break;
                if(strpointcount.ToLong(&pointcount) && pointcount >= 0 && pointcount <= 255)
                    logNotice(format_turn_config(name, pointcount));
            }
        }
        else if(type == wxT("LOG"))
        {
            // LOG <message>
            // Writes a message to the game log.
            logNotice(msg);
        }
        else if(type == wxT("CLEARLOG"))
        {
            // CLEARLOG
            // Erases the game log.
            clearLog();
            sender->respondgamerequest(getGameID(), wxT("CLEARLOG DONE"));
        }
        else if(type == wxT("BEGINTURN"))
        {
            // BEGINTURN <suspect> <diceroll>
            // Begins a suspect's turn.
            unsigned long n = 256, roll = 256;
            if(nextParam(msg).ToULong(&n) && n < 256 && nextParam(msg).ToULong(&roll) && roll < 256)
            {
                board.commitMove(active);
                Player *pl = getPlayer(active);
                if(pl != NULL)
                    pl->roomWarped = FALSE;
                dicepoints = diceroll = roll;
                sSuspect = board.suspectMax();
                sWeapon = board.weaponMax();
                sRoom = board.roomMax();
                sDisprove = 255;
                disprover = board.suspectMax();
                active = n;
                sender->respondgamerequest(getGameID(), wxT("BEGINTURN DONE"));
                update();
                postStatus();
            }
            else
            {
                sender->respondgamerequest(getGameID(), wxT("BEGINTURN FAIL"));
            }
        }
        else if(type == wxT("MOVE"))
        {
            // MOVE <x> <y> <room> <dicepoints>
            // Moves the current suspect; dicepoints is the number of dice points remaining, NOT the number used!
            unsigned long x = 256, y = 256, room = 256, points = 256;
            if(nextParam(msg).ToULong(&x) && x < 256 && nextParam(msg).ToULong(&y) && y < 256 &&
                    nextParam(msg).ToULong(&room) && room < 256 && nextParam(msg).ToULong(&points) &&
                    points < 256 && board.setSuspectPosition(active, x, y))
            {
                dicepoints = points;
                board.setRoomAt(x, y, room);
                sender->respondgamerequest(getGameID(), wxT("MOVE DONE"));
                update();
            }
            else
            {
                sender->respondgamerequest(getGameID(), wxT("MOVE FAIL"));
            }
        }
        else if(type == wxT("SUGGEST"))
        {
            // SUGGEST <suspect> <weapon> <room> <x> <y> <dicepoints>
            // Sets the current suggestion, and room-warps the suspect to the specified position.
            // Again, the dice points remaining are specified.
            unsigned long s = 256, w = 256, r = 256, x = 256, y = 256, points = 256;
            if(nextParam(msg).ToULong(&s) && s >= board.suspectMin() && s < board.suspectMax() &&
                    nextParam(msg).ToULong(&w) && w >= board.weaponMin() && w < board.weaponMax() &&
                    nextParam(msg).ToULong(&r) && r >= board.roomMin() && r < board.roomMax() &&
                    nextParam(msg).ToULong(&x) && nextParam(msg).ToULong(&y) &&
                    nextParam(msg).ToULong(&points) && points < 256 && board.commitMove(active))
            {
                board.commitMove(active);
                if(s != active && x < board.getWidth() && y < board.getHeight())
                {
                    board.setRoomAt(x, y, r);
                    board.setSuspectPosition(s, x, y);
                    board.commitMove(s);
                    Player *pl = getPlayer(s);
                    if(pl != NULL)
                        pl->roomWarped = TRUE;
                }
                sSuspect = s;
                sWeapon = w;
                sRoom = r;
                sDisprove = 255;
                disprover = board.suspectMax();
                dicepoints = points;
                sender->respondgamerequest(getGameID(), wxT("SUGGEST DONE"));
                update();
                postStatus();
            }
            else
            {
                sender->respondgamerequest(getGameID(), wxT("SUGGEST FAIL"));
            }
        }
        else if(type == wxT("DISPROVE"))
        {
            // DISPROVE <card>
            // Sets the card being shown to disprove the suggestion.
            // This is sent to the channel without the card parameter.
            // At the moment, this is implemented by setting the shown card to be the suspect if not specified.
            if(msg.empty())
            {
                sender->respondgamerequest(getGameID(), wxT("DISPROVE DONE"));
                if(sDisprove == 255)
                {
                    sDisprove = sSuspect;
                    update();
                    postStatus();
                }
            }
            else
            {
                unsigned long card = 256;
                if(nextParam(msg).ToULong(&card) && (card == sSuspect || card == sWeapon || card == sRoom))
                {
                    sDisprove = card;
                    sender->respondgamerequest(getGameID(), wxT("DISPROVE DONE"));
                    update();
                    postStatus();
                }
                else
                {
                    sender->respondgamerequest(getGameID(), wxT("DISPROVE FAIL"));
                }
            }
        }
        else if(type == wxT("ACCUSE"))
        {
            // ACCUSE <suspect> <weapon> <room> <playing> <dicepoints>
            // Sets the current accusation, and sets the current player's accusation state to <playing> (0 for incorrect, 2 for correct).
            unsigned long s = 256, w = 256, r = 256, playing = 256, points = 256;
            if(nextParam(msg).ToULong(&s) && s >= board.suspectMin() && s < board.suspectMax() &&
                    nextParam(msg).ToULong(&w) && w >= board.weaponMin() && w < board.weaponMax() &&
                    nextParam(msg).ToULong(&r) && r >= board.roomMin() && r < board.roomMax() &&
                    nextParam(msg).ToULong(&playing) && playing < 256 &&
                    nextParam(msg).ToULong(&points) && points < 256 && board.commitMove(active))
            {
                board.commitMove(active);
                sSuspect = s;
                sWeapon = w;
                sRoom = r;
                dicepoints = points;
                disprover = 255;
                Player *pl = getPlayer(active);
                if(pl != NULL)
                    pl->playing = playing;
                sender->respondgamerequest(getGameID(), wxT("ACCUSE DONE"));
                update();
                postStatus();
            }
            else
            {
                sender->respondgamerequest(getGameID(), wxT("ACCUSE FAIL"));
            }
        }
        else if(type == wxT("REVEAL"))
        {
            // REVEAL <disprover> <candisprove>
            // Sets the current disprover.
            unsigned long disprovr = 256, candisprov = 2;
            if(nextParam(msg).ToULong(&disprovr) && nextParam(msg).ToULong(&candisprov) && candisprov < 2 && reveal(disprovr, candisprov))
                sender->respondgamerequest(getGameID(), wxT("REVEAL DONE"));
            else
                sender->respondgamerequest(getGameID(), wxT("REVEAL FAIL"));
        }
        else
        {
            sender->respondgamerequest(getGameID(), type + wxT(" FAIL"));
        }
    }
}

void Game::removeNetUser(const NetUser *usr)
{
    if(usr == NULL)
        return;
    if(players != NULL)
    {
        bool lostplayer = FALSE;
        for(int n = 0; n < board.suspectCount(); ++n)
        {
            if(players[n].netuser == usr)
            {
                if(controlling)
                {
                    if(players[n].playing == 1)
                        lostplayer = TRUE;
                    players[n].netuser = NULL;
                    players[n].type = playerTypeFallback;
                    if(n == getCurrentPlayer() && players[n].isAI())
                        timer->Start();
                }
                else
                {
                    players[n].netuser = ((usr == controller) ? NULL : controller);
                    players[n].type = playerTypeFallback;
                }
            }
        }
        if(lostplayer && mgr != NULL)
            mgr->playerLost(this);
    }
    if(!controlling && controller == usr)
    {
        gamestate = 3;
        stopGame();
        controller = NULL;
    }
}

#include <wx/log.h>

bool Game::netStartGame(const NetUser *usr) const
{
    if(channel != NULL && controlling)
    {
        // Send board dimensions and card dealings to players.
        wxString request;
        wxString name;
        request << wxT("BEGINGAME 1 ") << (int) board.getWidth() << wxT(" ") << (int) board.getHeight();
        request << wxT(" ") << board.getNameSuspect() << wxT(" ") << board.suspectCount();
        request << wxT(" ") << board.getNameWeapon() << wxT(" ") << board.weaponCount();
        request << wxT(" ") << board.getNameRoom() << wxT(" ") << board.roomCount();
        request << wxT(" ") << (int) board.getSuggestTurns() << wxT(" ") << (int) board.getAccuseTurns();
        request << wxT(" ") << (int) board.imageWidth << wxT(" ") << (int) board.imageHeight << wxT(" ");
        wxFileName filename;
        filename = board.imageFile;
        request << filename.GetFullName();
        if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
            return FALSE;
        wxUint8 card = 0;
        while(card < board.cardCount())
        {
            request = wxT("CARDNAMES ");
            request << (int) card;
            while(card < board.cardCount() && request.size() < channel->getMaxGameRequestLength(getGameID()))
            {
                name = board.getCardName(card);
                name.Replace(wxT(" "), wxT("_"));
                request << wxT(" ") << name;
                ++card;
            }
            if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                return FALSE;
        }
        wxUint8 n = board.suspectMin();
        while(n < board.suspectMax())
        {
            request = wxT("SUSPECTPOS ");
            request << (int) n;
            while(n < board.suspectMax() && request.size() < 400)
            {
                request << wxT(" ") << (int) board.getOriginalX(n) << wxT(" ") << (int) board.getOriginalY(n);
                ++n;
            }
            if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                return FALSE;
        }
        if(!netSendCards(usr))
            return FALSE;
        request = wxT("LOGSETTINGS");
        if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
            return FALSE;
        for(n = 0; n < board.getTurnConfigCount(); ++n)
        {
            request = wxT("POINTCONF ");
            request << (int) n;
            name = board.getTurnConfigName(n);
            name.Replace(wxT(" "), wxT("_"));
            request << wxT(" ") << name << wxT(" ") << (int) board.getTurnConfigValue(n);
            if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                return FALSE;
        }
        // Set up the current game state.
        // FIXME: This needs to be tested and debugged.
        if(inProgress())
        {
#if 0
            // Erase the game log if required.
            if(autoClearLog)
            {
                request = wxT("CLEARLOG");
                if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                    return FALSE;
            }
#endif
            // Begin the turn.
            request = wxT("BEGINTURN ");
            request << (int) active << wxT(" ") << (int) diceroll;
            if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                return FALSE;
            // Perform the player's move.
            request = wxT("MOVE ");
            request << (int) board.getSuspectX(active) << wxT(" ") << (int) board.getSuspectY(active) << wxT(" ");
            request << (int) board.getSuspectRoom(active) << wxT(" ") << (int) dicepoints;
            if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                return FALSE;
            // Set up suggestion or accusation.
            if(hasSuggestion())
            {
                request = wxT("SUGGEST ");
                request << (int) sSuspect << wxT(" ") << (int) sWeapon << wxT(" ") << (int) sRoom << wxT(" ");
                request << (int) board.getSuspectX(sSuspect) << wxT(" ") << (int) board.getSuspectY(sSuspect) << wxT(" ");
                request << (int) dicepoints << wxT(" ");
                int holder = (active + 1) % board.suspectCount();
                while(holder != active && cards[sSuspect] != holder && cards[sWeapon] != holder && cards[sRoom] != holder)
                    holder = (holder + 1) % board.suspectCount();
                request << (int) holder;
                if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                    return FALSE;
            }
            else if(hasAccusation())
            {
                request = wxT("ACCUSE ");
                request << (int) sSuspect << wxT(" ") << (int) sWeapon << wxT(" ") << (int) sRoom << wxT(" ");
                if(cards[sSuspect] < board.suspectMax() || cards[sWeapon] < board.suspectMax() || cards[sRoom] < board.suspectMax())
                    request << (int) 0;
                else
                    request << (int) 2;
                request << wxT(" ") << (int) dicepoints;
                if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                    return FALSE;
            }
            // Reveal to the correct stage.
            if(disprover < board.suspectMax())
            {
                request = wxT("REVEAL ");
                request << ((int) disprover) << wxT(" ") << ((int) canDisprove());
                if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                    return FALSE;
            }
            // Indicate the disproof.
            if(sDisprove < board.cardCount())
            {
                request = wxT("DISPROVE");
                const Player *plactive = getPlayer(active);
                const Player *pldisprove = getPlayer(disprover);
                if(usr != NULL && ((plactive != NULL && plactive->netuser == usr) | (pldisprove != NULL && pldisprove->netuser == usr)))
                    request << wxT(" ") << (int) sDisprove;
                if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                    return FALSE;
            }
        }
        if(paused)
        {
            request = wxT("PAUSE");
            if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                return FALSE;
        }
    }
    return TRUE;
}

bool Game::netSendCards(const NetUser *usr) const
{
    if(channel != NULL && controlling)
    {
        const Player *pl;
        wxString request;
        wxUint8 n;
        wxUint8 card;
        for(n = board.suspectMin(); n < board.suspectMax(); ++n)
        {
            pl = getPlayer(n);
            if(pl != NULL && pl->type != 0)
            {
                request = wxT("SETPLAYER ");
                request << (int) n;
                request << wxT(" ") << (int) pl->type;
                if(pl->netuser == NULL)
                {
                    request << wxT(" ") << channel->getUser()->nick << wxT(" ") << pl->name;
                    if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                        return FALSE;
                }
                else
                {
                    request << wxT(" ") << pl->netuser->nick << wxT(" ") << pl->name;
                    if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
                        return FALSE;
                    if(usr == NULL || pl->netuser == usr)
                    {
                        request = wxT("HAVECARDS ");
                        request << (int) n;
                        for(card = 0; card < board.cardCount(); ++card)
                        {
                            if(cards[card] == n)
                                request << wxT(" ") << (int) card;
                        }
                        if(!pl->netuser->sendgamerequest(getGameID(), request))
                            return FALSE;
                    }
                }
            }
        }
        // Send settings.
        wxString name;
        request = wxT("SETTINGS");
        name = wxT(PACKAGE_STRING);
        name.Replace(wxT(" "), wxT("_"));
        request << wxT(" HOSTVER ") << name;
        name = board.getTitle();
        name.Replace(wxT(" "), wxT("_"));
        request << wxT(" BOARDTITLE ") << name;
        name = board.getDiceString();
        name.Replace(wxT(" "), wxT("_"));
        request << wxT(" DICE ") << name;
        request << wxT(" POINTCONFCOUNT ") << ((int) board.getTurnConfigCount());
        request << (roomWarp ? wxT(" ROOMWARP 1") : wxT(" ROOMWARP 0"));
        request << (autoClearLog ? wxT(" AUTOCLEARLOG 1") : wxT(" AUTOCLEARLOG 0"));
        request << (startSuggest ? wxT(" STARTSUGGEST 1") : wxT(" STARTSUGGEST 0"));
        request << (startSuggestWarped ? wxT(" STARTSUGGESTWARPED 1") : wxT(" STARTSUGGESTWARPED 0"));
        request << wxT(" END");
        if(!((usr == NULL) ? channel->sendgamerequest(getGameID(), request) : usr->sendgamerequest(getGameID(), request)))
            return FALSE;
    }
    return TRUE;
}

wxString Game::getGameDetails() const
{
    wxString desc;
    if(channel == NULL)
    {
        desc << wxT("This is a local game.\n");
    }
    else
    {
        desc << wxT("This game is being hosted");
        if(controlling)
        {
            if(channel->getUser() != NULL)
                desc << wxT(" by ") << channel->getUser()->nick;
            desc << wxT(" with " PACKAGE_STRING);
        }
        else if(controller != NULL)
        {
            desc << wxT(" by ") << controller->nick;
            if(!controllerver.empty())
                desc << wxT(" with ") << controllerver;
        }
        else
        {
            desc << wxT(" by an unknown user");
        }
        desc << wxT(" in ") << channel->name << wxT(".\n");
    }
    desc << wxT("Game type: ") << board.getTitle() << wxT("\n");
    if(roomWarp)
        desc << wxT("The suspect named in a suggestion will be moved into the room.\n");
    else
        desc << wxT("The suspect named in a suggestion will not be moved into the room.\n");
    if(autoClearLog)
        desc << wxT("The game log will be automatically cleared at the end of each turn.\n");
    else
        desc << wxT("The game log will show the entire game history.\n");
    if(startSuggest)
        desc << wxT("A player may suggest without moving");
    else
        desc << wxT("A player must move before suggesting");
    if(roomWarp && startSuggestWarped != startSuggest)
        desc << wxT(" unless they have been named in a suggestion.\n");
    else
        desc << wxT(".\n");
    desc << wxT("Movement points per turn: ") << board.getDiceString() << wxT("\n");
    for(wxUint8 n = 0; n < board.getTurnConfigCount(); ++n)
        desc << format_turn_config(board.getTurnConfigName(n), board.getTurnConfigValue(n)) << wxT("\n");
    return desc;
}

wxString Game::getStatusText() const
{
//    wxString status = channel == NULL ? wxT("Local game - ") : (wxT("Network game in ") + channel->name + wxT(" - "));
    wxString status;
    const Player *pl = getPlayer(active);
    if(pl == NULL)
    {
        if(gamestate > 1)
            status += wxT("Game aborted.");
        else
            status += wxT("Game initialising.");
    }
    else if(paused)
    {
        status += wxT("Game paused.");
    }
    else if(hasSuggestion())
    {
        if(sDisprove < board.cardCount())
        {
            status += /*wxT(" - ") +*/ board.getCardName(disprover) + wxT(" shows ") + board.getCardName(active);
            if(canControl(active) || canControl(disprover))
            {
                if(needHideCurrent())
                {
                    status += wxT(" a card.");
                    if(canControl(active))
                        status += wxT(" Use View Card to see it.");
                }
                else
                {
                    status += wxT(" the card [") + board.getCardName(sDisprove) + wxT("]");
                }
            }
            else
            {
                status += wxT(" a card.");
            }
        }
        else if(disprover == active)
        {
            status += wxT(/*" - "*/"The suggestion cannot be disproved.");
        }
        else if(disprover < board.suspectMax())
        {
            status += /*wxT(" - ") +*/ board.getCardName(disprover) + (canDisprove() ? wxT(" can") : wxT(" cannot")) + wxT(" disprove the suggestion.");
        }
        else
        {
            status += board.getCardName(active) + wxT(" suggests");
            status += wxT(" [") + board.getCardName(sSuspect) + wxT("]");
            status += wxT(" [") + board.getCardName(sWeapon) + wxT("]");
            status += wxT(" [") + board.getCardName(sRoom) + wxT("]");
        }
    }
    else if(hasAccusation())
    {
        status += board.getCardName(active);
        if(disprover < board.suspectMax())
        {
            status += (pl->playing == 2 ? wxT(" has accused correctly, and wins the game!")
                : wxT(" has accused incorrectly, and is out of the game!"));
        }
        else
        {
            status += wxT(" accuses [") + board.getCardName(sSuspect) + wxT("]");
            status += wxT(" [") + board.getCardName(sWeapon) + wxT("]");
            status += wxT(" [") + board.getCardName(sRoom) + wxT("]");
        }
    }
    else if(pl->playing == 1)
    {
        if(pl->netuser == NULL && pl->type == 1)
            status += board.getCardName(active) + wxT(", it is your turn! ");
        else
            status += wxT("It is ") + board.getCardName(active) + wxT("'s turn. ");
        if(dicepoints == 1)
            status += wxT("1 movement point available.");
        else
            status << (int) dicepoints << wxT(" movement points available.");
    }
    else
    {
        status += pl->playing == 0 ? wxT("Game drawn, solution was:") : wxT("Game won by ") + board.getCardName(active) + wxT(", solution was:");
        for(wxUint8 card = 0; card < board.cardCount(); ++card)
        {
            if(cards[card] == board.suspectMax())
                status += wxT(" [") + board.getCardName(card) + wxT("]");
        }
    }
    return status;
}

wxString Game::getShortStatus() const
{
    const Player *pl = getPlayer(active);
    switch(gamestate)
    {
    case 0:
        return wxT("Not Started");
    case 1:
        return paused ? wxT("Paused") : wxT("In Progress");
    case 2:
        if(pl == NULL)
            return wxT("Finished");
        return (pl->playing == 0) ? wxT("Drawn") : (wxT("Won by ") + board.getCardName(active));
    case 3:
        return wxT("Aborted");
    }
    return wxT("Unknown");
}

void Game::setDescription(const wxString &newval, bool sendnet)
{
    if(newval != description)
    {
        description = newval;
        if(sendnet)
            updateTableTopic();
    }
}

wxUint8 Game::getPlayerCount() const
{
    wxUint8 playerCount = 0;
    const Player *player;
    for(wxUint8 suspect = board.suspectMin(); suspect < board.suspectMax(); ++suspect)
    {
        player = getPlayer(suspect);
        if(player != NULL && player->type > 0)
            ++playerCount;
    }
    return playerCount;
}

bool Game::dealCards()
{
    if(cards == NULL || players == NULL || board.suspectCount() == 0)
        return FALSE;
    // Initialise the pack of cards.
    wxUint8 packsize = board.cardCount();
    wxUint8 pack[packsize];
    wxUint8 card;
    for(card = 0; card < packsize; ++card)
        pack[card] = card;
    // Choose "murder cards".
    wxUint8 thesuspect = board.suspectMin() + std::rand() / (RAND_MAX / board.suspectCount() + 1);
    wxUint8 theweapon = board.weaponMin() + std::rand() / (RAND_MAX / board.weaponCount() + 1);
    wxUint8 theroom = board.roomMin() + std::rand() / (RAND_MAX / board.roomCount() + 1);
    cards[thesuspect] = cards[theweapon] = cards[theroom] = board.suspectCount();
    // Remove these cards from the pack.
    // We remove cards by moving the last pack entry into the slot of the card to be removed and decreasing the pack size.
    // Here we can safely refer to specific cards as long as we do not remove a card of higher index before one with a lower index.
    pack[theroom] = pack[--packsize];
    pack[theweapon] = pack[--packsize];
    pack[thesuspect] = pack[--packsize];
    // Get list of enabled players.
    wxUint8 playercount = 0;
    wxUint8 playertosuspect[board.suspectCount()];
    // Humans first.
    for(card = 0; card < board.suspectCount(); ++card)
    {
        if(players[card].isHuman())
        {
            playertosuspect[playercount] = card;
            ++playercount;
        }
    }
    // Then AIs.
    for(card = 0; card < board.suspectCount(); ++card)
    {
        if(players[card].isAI())
        {
            playertosuspect[playercount] = card;
            ++playercount;
        }
    }
    if(playercount == 0)
        return FALSE;
    // Until the pack is empty, pick a random card from the pack, remove it and give it to the next player.
    wxUint8 player = 0;
    while(packsize > 0) {
        card = std::rand() / (RAND_MAX / packsize + 1);
        cards[pack[card]] = playertosuspect[player];
        pack[card] = pack[--packsize];
        player = (player + 1) % playercount;
    }
    return TRUE;
}

bool Game::canControl(wxUint8 n, bool needplaying) const
{
    if(!board.hasLocations())
        return FALSE;
    const Player *player = getPlayer(n);
    return player != NULL && player->type == 1 && player->netuser == NULL && (!needplaying || player->playing == 1);
}

wxUint8 Game::getControllableCount(bool needplaying) const
{
    wxUint8 count = 0;
    for(wxUint8 n = board.suspectMin(); n < board.suspectMax(); ++n)
    {
        if(canControl(n, needplaying))
            ++count;
    }
    return count;
}

bool Game::needHideCurrent() const
{
    for(wxUint8 n = board.suspectMin(); n < board.suspectMax(); ++n)
    {
        if(n != active && n != disprover && canControl(n))
            return TRUE;
    }
    return FALSE;
}

bool Game::move(wxUint8 x, wxUint8 y)
{
    if(x == board.getSuspectX(active) && y == board.getSuspectY(active))
        return TRUE;
    if(!controlling)
    {
        if(controller == NULL)
            return FALSE;
        wxString request;
        request << wxT("MOVE ") << (int) x << wxT(" ") << (int) y;
        return controller->sendgamerequest(getGameID(), request);
    }
    // Moving is not allowed after a suggestion or accusation.
    if(hasSuggestion() || hasAccusation())
        return FALSE;
    // Just cancel the move if it is for the square on which the character started.
    if(x == board.getOriginalX(active) && y == board.getOriginalY(active))
        return cancelMove();
    // Save old location.
    wxUint8 oldx = board.getSuspectX(active);
    wxUint8 oldy = board.getSuspectY(active);
    // Get number of dice points required.
    wxUint8 turns = board.getMoveTurns(board.getSuspectX(active), board.getSuspectY(active), x, y, dicepoints, &x, &y);
    if(turns > dicepoints || !board.setSuspectPosition(active, x, y))
        return FALSE;
    // Ensure that the suspect cannot leave a room on the same turn as entering it.
    wxUint8 newroom = board.getSuspectRoom(active);
    if(newroom != board.getRoomAt(oldx, oldy))
    {
        if(changedRooms)
        {
            board.setSuspectPosition(active, oldx, oldy);
            return FALSE;
        }
        changedRooms = board.getSuspectRoom(active) < board.roomMax();
    }
    // Deduct the required dice points - or reset if back to starting position.
    if(board.isMoving(active))
        dicepoints -= turns;
    else
        dicepoints = diceroll;
    // Tell the network players about the move.
    if(channel != NULL)
    {
        wxString request;
        request << wxT("MOVE ") << (int) x << wxT(" ") << (int) y;
        request << wxT(" ") << (int) board.getSuspectRoom(active) << wxT(" ") << (int) dicepoints;
        if(!channel->sendgamerequest(getGameID(), request))
            return FALSE;
    }
    return TRUE;
}

bool Game::cancelMove()
{
    if(!controlling)
    {
        if(controller == NULL)
            return FALSE;
        return controller->sendgamerequest(getGameID(), wxT("CANCELMOVE"));
    }
    changedRooms = FALSE;
    if(!resetDice())
        return FALSE;
    if(!board.cancelMove(active))
        return FALSE;
    if(channel != NULL)
    {
        wxString request;
        request << wxT("MOVE ") << (int) board.getSuspectX(active) << wxT(" ") << (int) board.getSuspectY(active);
        request << wxT(" ") << (int) board.getSuspectRoom(active) << wxT(" ") << (int) dicepoints;
        if(!channel->sendgamerequest(getGameID(), request))
            return FALSE;
    }
    return TRUE;
}

bool Game::canSuggest() const
{
    if(hasSuggestion() || hasAccusation() || board.getSuggestTurns() > dicepoints)
        return FALSE;
    if(board.getSuspectRoom(active) < board.roomMin() || board.getSuspectRoom(active) >= board.roomMax())
        return FALSE;
    if(board.getSuspectRoom(active) != board.getOriginalRoom(active))
        return TRUE;
    const Player *pl = getPlayer(active);
    return (pl != NULL && pl->roomWarped) ? startSuggestWarped : startSuggest;
}

bool Game::suggest(wxUint8 suspect, wxUint8 weapon)
{
    if(!controlling)
    {
        if(controller == NULL)
            return FALSE;
        wxString request;
        request << wxT("SUGGEST ") << (int) suspect << wxT(" ") << (int) weapon;
        return controller->sendgamerequest(getGameID(), request);
    }
    if(!canSuggest())
        return FALSE;
    // Get room.
    wxUint8 room = board.getSuspectRoom(active);
    // Checks on card ranges.
    if(suspect < board.suspectMin() || suspect >= board.suspectMax()
      || weapon < board.weaponMin() || weapon >= board.weaponMax()
      || room < board.roomMin() || room >= board.roomMax())
        return FALSE;
    // Commit the player's board movement.
    if(!board.commitMove(active))
        return FALSE;
    // If another suspect is suggested, they should be called into the room.
    if(roomWarp && suspect != active)
    {
        if(!board.roomWarp(suspect, room))
            return FALSE;
        Player *pl = getPlayer(suspect);
        if(pl != NULL)
            pl->roomWarped = TRUE;
    }
    // Set suggestion cards.
    sSuspect = suspect;
    sWeapon = weapon;
    sRoom = room;
    sDisprove = 255;
    disprover = board.suspectMax();
    // Deduct dice points.
    dicepoints -= board.getSuggestTurns();
    // Start timer for disproving.
    if(!timer->Start())
        return FALSE;
    // Inform players in network game.
    if(channel != NULL)
    {
        wxString request;
        request << wxT("SUGGEST ") << (int) sSuspect << wxT(" ") << (int) sWeapon << wxT(" ") << (int) sRoom;
        if(roomWarp)
            request << wxT(" ") << (int) board.getSuspectX(sSuspect) << wxT(" ") << (int) board.getSuspectY(sSuspect);
        else
            request << wxT(" ") << (int) board.getWidth() << wxT(" ") << (int) board.getHeight();
        request << wxT(" ") << (int) dicepoints;
        int holder = (active + 1) % board.suspectCount();
        while(holder != active && cards[sSuspect] != holder && cards[sWeapon] != holder && cards[sRoom] != holder)
            holder = (holder + 1) % board.suspectCount();
        request << wxT(" ") << (int) holder;
        if(!channel->sendgamerequest(getGameID(), request))
            return FALSE;
    }
    postStatus();
    return TRUE;
}

bool Game::canDisprove() const
{
    if(!hasSuggestion() || sDisprove < board.cardCount() || disprover < board.suspectMin() || disprover >= board.suspectMax() || disprover == active)
        return FALSE;
    if(!controlling)
        return candisprove;
    return cards[sSuspect] == disprover || cards[sWeapon] == disprover || cards[sRoom] == disprover;
}

bool Game::disprove(wxUint8 card)
{
    if(!controlling)
    {
        if(controller == NULL)
            return FALSE;
        wxString request;
        request << wxT("DISPROVE ") << (int) card;
        return controller->sendgamerequest(getGameID(), request);
    }
    if(!canDisprove())
        return FALSE;
    if(card != sSuspect && card != sWeapon && card != sRoom)
        return FALSE;
    if(cards[card] != disprover)
        return FALSE;
    sDisprove = card;
    if(channel != NULL)
    {
        wxString request;
        request << wxT("DISPROVE ") << (int) sDisprove;
        Player *plactive = getPlayer(active);
        if(plactive != NULL && plactive->netuser != NULL && !plactive->netuser->sendgamerequest(getGameID(), request))
            return FALSE;
        Player *pldisprove = getPlayer(disprover);
        if(pldisprove != NULL && pldisprove->netuser != NULL && pldisprove->netuser != plactive->netuser &&
          !pldisprove->netuser->sendgamerequest(getGameID(), request))
            return FALSE;
        if(!channel->sendgamerequest(getGameID(), wxT("DISPROVE")))
            return FALSE;
    }
    postStatus();
    return TRUE;
}

bool Game::canAccuse() const
{
    return !activeSuggestion() && !hasAccusation() && board.getAccuseTurns() <= dicepoints;
}

bool Game::accuse(wxUint8 suspect, wxUint8 weapon, wxUint8 room)
{
    if(!controlling)
    {
        if(controller == NULL)
            return FALSE;
        wxString request;
        request << wxT("ACCUSE ") << (int) suspect << wxT(" ") << (int) weapon << wxT(" ") << (int) room;
        return controller->sendgamerequest(getGameID(), request);
    }
    if(!canAccuse())
        return FALSE;
    // Checks on card ranges.
    if(suspect < board.suspectMin() || suspect >= board.suspectMax()
      || weapon < board.weaponMin() || weapon >= board.weaponMax()
      || room < board.roomMin() || room >= board.roomMax())
        return FALSE;
    // Commit the player's board movement.
    if(!board.commitMove(active))
        return FALSE;
    // Set accusation cards.
    sSuspect = suspect;
    sWeapon = weapon;
    sRoom = room;
    disprover = board.suspectMax();
    // Deduct dice points.
    dicepoints -= board.getAccuseTurns();
    // Make the player win or lose as appropriate.
    Player *pl = getPlayer(active);
    wxUint8 playing = (cards[sSuspect] < board.suspectMax() || cards[sWeapon] < board.suspectMax() || cards[sRoom] < board.suspectMax()) ? 0 : 2;
    if(pl != NULL)
        pl->playing = playing;
    // Start timer for revealing result.
    if(!timer->Start())
        return FALSE;
    // Tell network players.
    if(channel != NULL)
    {
        wxString request;
        request << wxT("ACCUSE ") << (int) sSuspect << wxT(" ") << (int) sWeapon << wxT(" ") << (int) sRoom;
        request << wxT(" ") << (int) playing << wxT(" ") << (int) dicepoints;
        if(!channel->sendgamerequest(getGameID(), request))
            return FALSE;
    }
    postStatus();
    return TRUE;
}

bool Game::needUserInput() const
{
    if(!inProgress())
        return FALSE;
    if(hasAccusation())
        return FALSE;
    if(!hasSuggestion() || sDisprove < board.cardCount() || disprover == active)
        return canControl(active);
    return canControl(disprover) && canDisprove();
}

bool Game::startGame()
{
    active = 255;
    return dealCards() && netStartGame() && logNotice(getGameDetails(), TRUE, FALSE) && startTurn();
}

bool Game::startTurn()
{
    timer->Stop();
    Player *pl = getPlayer(active);
    if(pl != NULL)
    {
//        if(pl->playing != 1)
//            return FALSE;
        pl->roomWarped = FALSE;
    }
    if(active < board.suspectMax() && !board.commitMove(active))
        return FALSE;
    if(!rollDice())
        return FALSE;
    changedRooms = FALSE;
    sSuspect = board.suspectMax();
    sWeapon = board.weaponMax();
    sRoom = board.roomMax();
    sDisprove = 255;
    disprover = board.suspectMax();
    if(autoClearLog && active < board.suspectMax())
        clearLog();
    wxUint8 n = board.suspectMin();
    do
    {
        if(++n > board.suspectMax())
            return FALSE;
        active = (active >= board.suspectMax() - 1) ? board.suspectMin() : (active + 1);
        pl = getPlayer(active);
    } while(pl == NULL || pl->type == 0 || pl->playing != 1);
    if(pl != NULL && pl->playing == 1 && gamestate != 1)
    {
        gamestate = 1;
        updateTableTopic();
    }
    postStatus();
    if(channel != NULL)
    {
        wxString request;
        if(autoClearLog)
        {
            request = wxT("CLEARLOG");
            if(!channel->sendgamerequest(getGameID(), request))
                return FALSE;
        }
        request = wxT("BEGINTURN ");
        request << (int) active << wxT(" ") << (int) diceroll;
        if(!channel->sendgamerequest(getGameID(), request))
            return FALSE;
    }
    if(pl != NULL && pl->isAI() && !timer->Start())
        return FALSE;
    return TRUE;
}

bool Game::endTurn()
{
    if(!controlling)
    {
        if(controller == NULL)
            return FALSE;
        return controller->sendgamerequest(getGameID(), wxT("ENDTURN"));
    }
    return startTurn();
}

void Game::update() const
{
    if(mgr != NULL)
        mgr->update(this);
}

void Game::postStatus(bool alert)
{
    wxString status = getStatusText();
    size_t pos = status.rfind(wxT(" - "));
    if(pos != wxString::npos)
        status = status.substr(pos + 3);
    status += wxT("\n");
    log += status;
    if(mgr != NULL)
    {
        mgr->addLog(this, status);
        if(alert)
        {
            while(status[status.size() - 1] == wxT('\n'))
                status.erase(status.size() - 1);
            if(canControl() && hasDisproved())
                status += wxT("\n\nDon't forget to click Accuse or End Turn when you are ready!");
            mgr->alert(this, status);
        }
    }
}

void Game::updateTableTopic() const
{
    if(controlling && channel != NULL)
    {
        Network *net = channel->getNetwork();
        if(net != NULL)
        {
            net->command(wxT("TOPIC ") + channel->name + wxT(" :Host: ") + net->getNick() + wxT(" :: Status: ") + getShortStatus() +
                wxT(" :::: ") + (description.empty() ? wxT("MurderSpot") : description));
        }
    }
}

bool Game::logNotice(wxString msg, bool showlocal, bool sendtonet, const NetUser *netuser)
{
    size_t pos;
    if(msg.empty())
        return TRUE;
    if(msg[msg.size() - 1] != wxT('\n'))
        msg += wxT("\n");
    if(showlocal)
    {
        log += msg;
        if(mgr != NULL)
            mgr->addLog(this, msg);
    }
    if(sendtonet && channel != NULL && controlling)
    {
        wxString request;
        while(!msg.empty())
        {
            pos = msg.find(wxT('\n'));
            request = wxT("LOG ") + msg.substr(0, pos);
            msg.erase(0, pos + 1);
            if(!((netuser == NULL) ? channel->sendgamerequest(getGameID(), request) : netuser->sendgamerequest(getGameID(), request)))
                return FALSE;
        }
    }
    return TRUE;
}

void Game::clearLog()
{
    log.clear();
    if(mgr != NULL)
        mgr->clearLog(this);
}

bool Game::reveal(wxUint8 disprovr, bool candisprov)
{
    if(hasSuggestion() && !hasDisproved() && !canDisprove())
    {
        Player *pl;
        if(disprovr < board.suspectMax())
        {
            disprover = disprovr;
            pl = getPlayer(disprover);
        }
        else
        {
            if(disprover >= board.suspectMax())
                disprover = active;
            wxUint8 n = board.suspectMin();
            do
            {
                if(++n > board.suspectMax())
                    return FALSE;
                disprover = (disprover >= board.suspectMax() - 1) ? board.suspectMin() : (disprover + 1);
                pl = getPlayer(disprover);
            } while(pl == NULL || pl->type == 0);
        }
        candisprove = candisprov;
        update();
        postStatus();
        if(disprover == active || canDisprove())
        {
            if(pl != NULL && pl->isAI() && !timer->Start())
                return FALSE;
        }
        else
        {
            if(!timer->Start())
                return FALSE;
        }
    }
    else if(hasAccusation() && gamestate == 1)
    {
        if(disprover == 0)
        {
            Player *pl = getPlayer(active);
            if(pl != NULL && pl->playing == 0)
            {
                startTurn();
            }
            else
            {
                sSuspect = 255;
                sWeapon = 255;
                sRoom = 255;
                disprover = 255;
            }
            if(!inProgress())
            {
                gamestate = 2;
                updateTableTopic();
                postStatus();
                if(channel != NULL && controlling)
                {
                    // Game has finished - assign murder cards so that players get correct display.
                    wxString request;
                    request << wxT("HAVECARDS ") << (int) (board.suspectMax());
                    for(wxUint8 card = 0; card < board.cardCount(); ++card)
                    {
                        if(cards[card] == board.suspectMax())
                            request << wxT(" ") << (int) card;
                    }
                    if(!channel->sendgamerequest(getGameID(), request))
                        return FALSE;
                }
            }
        }
        else
        {
            disprover = 0;
            if(!timer->Start())
                return FALSE;
            postStatus();
        }
        update();
    }
    else
    {
        return TRUE;
    }
    if(channel != NULL && controlling)
    {
        wxString request;
        request << wxT("REVEAL ") << (int) disprover << wxT(" ") << (int) canDisprove();
        if(!channel->sendgamerequest(getGameID(), request))
            return FALSE;
    }
    return TRUE;
}

bool Game::AIplay()
{
    if(hasAccusation())
        return TRUE;
    Player *pl = getPlayer(getCurrentPlayer());
    if(pl == NULL || !pl->isAI())
        return TRUE;
    if(canDisprove())
    {
        bool haveSuspect = TRUE, haveWeapon = TRUE, haveRoom = TRUE;
        while(haveSuspect || haveWeapon || haveRoom)
        {
            switch(std::rand() % 3)
            {
            case 0:
                if(disprove(sSuspect))
                {
                    update();
                    return TRUE;
                }
                haveSuspect = FALSE;
                break;
            case 1:
                if(disprove(sWeapon))
                {
                    update();
                    return TRUE;
                }
                haveWeapon = FALSE;
                break;
            case 2:
                if(disprove(sRoom))
                {
                    update();
                    return TRUE;
                }
                haveRoom = FALSE;
                break;
            default:
                return FALSE;
            }
        }
    }
    else if(!activeSuggestion())
    {
        if(endTurn())
        {
            update();
            return TRUE;
        }
    }
    return FALSE;
}

void GameTimer::Notify()
{
    if(game != NULL)
    {
        game->AIplay();
        game->reveal();
    }
}
