// MurderSpot - a free game based on Clue/Cluedo
// Copyright 2005-2009 Christopher Elsby <chrise@chrise.me.uk>
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef MURDERSPOT_GAME_H
#define MURDERSPOT_GAME_H

#include "board.h"
#include <wx/defs.h>
#include <wx/string.h>

class Network;
class NetUser;
class NetChannel;
class Manager;
class GameTimer;

const wxUint8 playerTypeCount = 3;
const wxUint8 playerTypeFallback = 2;
const wxChar * const playerTypeNames[playerTypeCount] = {wxT("None"), wxT("Human on this computer"), wxT("Computer Player")};

class Player
{
public:
    Player() : type(0), netuser(NULL), playing(1), roomWarped(FALSE) { }
    bool isHuman() const {return type == 1;}
    bool isAI() const {return type > 1;}
    // Name of the player.
    wxString name;
    // Type of player.
    wxUint8 type;
    // Controlling network user (NULL for local).
    const NetUser *netuser;
    // Accusation state: 0 = incorrect accusation made (out of game), 1 = no accusation made (playing), 2 = correct accusation made (winner).
    wxUint8 playing;
    // Whether the player has been called into a room by another's suggestion.
    bool roomWarped;
};

class Game
{
public:
    Game();
    virtual ~Game();
    // Returns a pointer to the board.
    Board *getBoard() {return &board;}
    const Board *getBoard() const {return &board;}
    // Indicates whether or not the game may be discarded (i.e. it has finished or been aborted).
    bool canDiscard() const {return gamestate > 1;}
    // Checks whether the game is in progress.
    bool inProgress() const {const Player *pl = getPlayer(active); return pl != NULL && (pl->playing == 1 || hasAccusation());}
    bool started() const {const Player *pl = getPlayer(active); return pl != NULL;}
    // Pauses the game.
    void pauseGame(bool pause = TRUE);
    void resumeGame() {pauseGame(FALSE);}
    bool isPaused() const {return paused;}
    // Stops the game.
    void stopGame(bool updateState = TRUE);
    // Resets the object.
    // This must be called whenever a new board is loaded.
    void reset();
    // The manager attached to the game.
    Manager *getManager() const {return mgr;}
    void setManager(Manager *manager) {mgr = manager;}
    // Network information.
    const NetChannel *getChannel() const {return channel;}
    void setChannel(const NetChannel *chan) {channel = chan;}
    Network *getNetwork() const;
    const NetUser *getController() const {return controller;}
    bool getControlling() const {return controlling;}
    void makeHost() {controlling = TRUE; controller = NULL;}
    void makeGuest(const NetUser *ctrler = NULL) {controlling = FALSE; controller = ctrler;}
    wxString getGameID() const;
    // Handles a request from network.
    void processNet(const NetUser *sender, wxString type, wxString msg);
    // Removes all pointers to the user object.
    void removeNetUser(const NetUser *usr);
    // Sends game initialisation messages to the network.
    // If a user is specified, only that user will be sent messages.
    bool netStartGame(const NetUser *usr = NULL) const;
    // Sends player assignment, card ownership and game settings messages.
    bool netSendCards(const NetUser *usr = NULL) const;
    // Returns a message describing the game.
    wxString getGameDetails() const;
    // Returns text to be displayed in the status bar.
    wxString getStatusText() const;
    // Returns text to be displayed in the table list.
    wxString getShortStatus() const;
    // The game description.
    wxString getDescription() const {return description;}
    void setDescription(const wxString &newval, bool sendnet = TRUE);
    // Returns the game log.
    const wxString &getLog() const {return log;}
    // Retrieves a player object.
    wxUint8 getPlayerCount() const;
    Player *getPlayer(wxUint8 suspect) {if(players == NULL || board.suspectCount() < 1 || suspect >= board.suspectMax()) return NULL; return players + suspect;}
    const Player *getPlayer(wxUint8 suspect) const {if(players == NULL || suspect >= board.suspectMax()) return NULL; return players + suspect;}
    // Card ownership.
    wxUint8 getCardHolder(wxUint8 card) const {if(cards == NULL || card >= board.cardCount()) return 255; return cards[card];}
    bool setCardHolder(wxUint8 card, wxUint8 suspect)
        {if(cards == NULL || card >= board.cardCount() || suspect > board.suspectMax()) return FALSE; cards[card] = suspect; return TRUE;}
    // Randomly assigns the cards.
    bool dealCards();
    // Returns whether the user can control a player.
    bool canControl(bool needplaying = TRUE) const {return canControl(active, needplaying);}
    bool canControl(wxUint8 n, bool needplaying = FALSE) const;
    wxUint8 getControllableCount(bool needplaying = FALSE) const;
    bool needHideCards() const {return getControllableCount() > 1;}
    bool needHideCurrent() const;
    // Dice and moves.
    wxUint8 getDiceRoll() const {return diceroll;}
    wxUint8 getDicePoints() const {return dicepoints;}
    bool rollDice() {diceroll = dicepoints = board.getRoll(); return TRUE;}
    bool useDicePoints(wxUint8 points) {if(points > dicepoints) return FALSE; dicepoints -= points; return TRUE;}
    bool resetDice() {dicepoints = diceroll; return TRUE;}
    bool move(wxUint8 x, wxUint8 y);
    bool commitMove() {return board.commitMove(active);}
    bool cancelMove();
    bool isMoving() const {return board.isMoving(active);}
    // Suggestions and accusations.
    bool hasSuggestion() const
        {const Player *pl = getPlayer(active); return pl != NULL && pl->playing == 1 &&
            sSuspect >= board.suspectMin() && sSuspect < board.suspectMax() &&
            sWeapon >= board.weaponMin() && sWeapon < board.weaponMax() &&
            sRoom >= board.roomMin() && sRoom < board.roomMax();}
    bool activeSuggestion() const
        {return hasSuggestion() && sDisprove >= board.cardCount() && disprover != active;}
    bool hasDisproved() const
        {return sDisprove < board.cardCount();}
    bool hasAccusation() const
        {const Player *pl = getPlayer(active); return pl != NULL && pl->playing != 1 &&
            sSuspect >= board.suspectMin() && sSuspect < board.suspectMax() &&
            sWeapon >= board.weaponMin() && sWeapon < board.weaponMax() &&
            sRoom >= board.roomMin() && sRoom < board.roomMax();}
    wxUint8 suggestedSuspect() const {return sSuspect;}
    wxUint8 suggestedWeapon() const {return sWeapon;}
    wxUint8 suggestedRoom() const {return sRoom;}
    wxUint8 disprovedCard() const {return sDisprove;}
    bool canSuggest() const;
    bool suggest(wxUint8 suspect, wxUint8 weapon);
    bool canDisprove() const;
    bool disprove(wxUint8 card);
    bool canAccuse() const;
    bool accuse(wxUint8 suspect, wxUint8 weapon, wxUint8 room);
    // Active player.
    wxUint8 getActive() const {return active;}
    bool setActive(wxUint8 suspect) {if(suspect >= board.suspectMax()) return FALSE; active = suspect; return TRUE;}
    wxUint8 getDisprover() const {return disprover;}
    bool setDisprover(wxUint8 suspect) {if(suspect >= board.suspectMax()) return FALSE; disprover = suspect; return TRUE;}
    wxUint8 getCurrentPlayer() const {return (disprover < board.suspectMax() && sDisprove >= board.cardCount()) ? disprover : active;}
    // Indicates whether we are waiting for the user to do something.
    bool needUserInput() const;
    // Turns.
    bool startGame();
    bool startTurn();
    bool endTurn();
    // Updates the game window.
    void update() const;
    // Posts a status message in chat.
    void postStatus() {postStatus(needUserInput());}
    void postStatus(bool alert);
    // Updates the status of the table.
    void updateTableTopic() const;
    // Posts a notice to the game log.
    bool logNotice(wxString msg, bool showlocal = TRUE, bool sendtonet = TRUE, const NetUser *netuser = NULL);
    // Erases the game log.
    void clearLog();
    // Advance suggestion/accusation displays.
    // disprovr is the suspect whose turn it is to attempt to disprove (only applicable to suggestions).
    bool reveal(wxUint8 disprovr = 255, bool candisprov = FALSE);
    // AI actions.
    bool AIplay();
    // Whether room warp should be used.
    bool roomWarp;
    // Whether the log should be cleared at the end of each turn.
    bool autoClearLog;
    // Whether a player need not move rooms before suggesting.
    bool startSuggest;
    // Whether a player need not move rooms before suggesting if they have been room warped since their last turn.
    bool startSuggestWarped;
private:
    // Game manager.
    Manager *mgr;
    // Whether the game is paused.
    bool paused;
    // Game status.
    // 0 = not initialised, 1 = normal, 2 = finished, 3 = aborted
    wxUint8 gamestate;
    // Current suggestion cards.
    wxUint8 sSuspect, sWeapon, sRoom;
    // Card being shown to disprove suggestion.
    wxUint8 sDisprove;
    // Suspect currently being asked to disprove.
    // This is used as a flag when there is an accusation.
    wxUint8 disprover;
    // Whether the disprover can disprove (only used when the guest of a network game).
    bool candisprove;
    // Game board.
    Board board;
    // Player controllers.
    Player *players;
    // Which player holds each card (board.suspectMax() is used to indicate the "murder cards").
    wxUint8 *cards;
    // Which player's turn it is.
    wxUint8 active;
    // Dice roll and points remaining.
    wxUint8 diceroll, dicepoints;
    // Whether the player has changed rooms yet this turn.
    bool changedRooms;
    // Network channel being used (NULL for local game).
    const NetChannel *channel;
    // A textual description of the game (for the table list).
    wxString description;
    // Whether we are controlling this game.
    bool controlling;
    // Network user in charge of the game.
    const NetUser *controller;
    // The version of MurderSpot in use by the controller.
    wxString controllerver;
    // Timer.
    GameTimer *timer;
    // Game log.
    wxString log;
};

#endif // !defined(MURDERSPOT_GAME_H)
