// MurderSpot - a free game based on Clue/Cluedo
// Copyright 2005-2009 Christopher Elsby <chrise@chrise.me.uk>
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "global.h"

#include "network.h"
#include "manager.h"
#include <string.h>
#include <wx/socket.h>
#include <wx/string.h>
#include <wx/utils.h>
#include <wx/listimpl.cpp>
#include <wx/strconv.h>

WX_DEFINE_LIST(NetUserChannelList);

static wxString nextIRCparam(wxString &msg, bool checkColon = TRUE)
{
    while(!msg.empty() && msg.at(0) == wxT(' '))
        msg.erase(0, 1);
    if(msg.empty())
        return wxEmptyString;
    wxString param;
    if(checkColon && msg.at(0) == wxT(':'))
    {
        param = msg.substr(1);
        msg.clear();
    }
    else
    {
        size_t pos = msg.find(wxT(" "));
        if(pos < wxString::npos)
        {
            param = msg.substr(0, pos);
            msg.erase(0, pos + 1);
        }
        else
        {
            param = msg;
            msg.clear();
        }
    }
    return param;
}

NetUserChannel *NetUser::getChannelData(const NetChannel *chan) const
{
    if(chan == NULL)
        return NULL;
    NetUserChannelList::Node *node;
    for(node = channels.GetFirst(); node != NULL; node = node->GetNext())
    {
        if(node->GetData() != NULL && node->GetData()->getChannel() == chan)
            return node->GetData();
    }
    return NULL;
}

#if 0
bool NetUser::canHost(const NetChannel *chan) const
{
    NetUserChannel *data = getChannelData(chan);
    if(data == NULL)
        return FALSE;
    return data->canHost();
}
#endif

wxString NetUser::getHostMask() const
{
    if(nick.empty() || !knowhostmask)
        return wxEmptyString;
    wxString mask = nick + wxT("!") + user;
    if(!host.empty())
        mask += wxT("@") + host;
    return mask;
}

NetUserChannel *NetChannel::getUserData(const NetUser *usr) const
{
    if(usr == NULL)
        return NULL;
    NetUserChannelList::Node *node;
    for(node = users.GetFirst(); node != NULL; node = node->GetNext())
    {
        if(node->GetData() != NULL && node->GetData()->getUser() == usr)
            return node->GetData();
    }
    return NULL;
}

Network::~Network()
{
    NetUserChannelList::Node *node;
    for(NetUserMap::iterator uit = users.begin(); uit != users.end(); ++uit)
    {
        if(uit->second != NULL && uit->second != user)
        {
            while((node = uit->second->channels.GetFirst()) != NULL)
                removeUserChannel(node->GetData());
            delete uit->second;
        }
    }
    if(user != NULL)
    {
        while((node = user->channels.GetFirst()) != NULL)
            removeUserChannel(node->GetData());
        delete user;
    }
    for(NetChannelMap::iterator cit = channels.begin(); cit != channels.end(); ++cit)
    {
        if(cit->second != NULL)
        {
            while((node = cit->second->users.GetFirst()) != NULL)
                removeUserChannel(node->GetData());
            delete cit->second;
        }
    }
}

const NetChannel *Network::getChannel(wxString name) const
{
    if(name.at(0) != wxT('#'))
        return NULL;
    NetChannelMap::const_iterator it = channels.find(name);
    if(it != channels.end())
        return it->second;
    // This does not actually alter what the object represents, but
    // would normally be prevented by const.
    NetChannel *chan = new NetChannel((Network*) this);
    chan->name = name;
    return ((Network*) this)->channels[name] = chan;
}

NetChannel *Network::fetchChannel(wxString name)
{
    NetChannelMap::iterator it = channels.find(name);
    if(it != channels.end())
        return it->second;
    NetChannel *chan = new NetChannel(this);
    chan->name = name;
    return channels[name] = chan;
}

const NetUser *Network::getUser(wxString name) const
{
    if(name.empty() || name.at(0) == wxT('#'))
        return NULL;
    NetUserMap::const_iterator it = users.find(name);
    if(it != users.end())
        return it->second;
    NetUser *usr = new NetUser((Network*) this);
    usr->nick = name;
    return ((Network*) this)->users[name] = usr;
}

NetUser *Network::fetchUser(wxString name)
{
    NetUserMap::iterator it = users.find(name);
    if(it != users.end())
        return it->second;
    NetUser *usr = new NetUser(this);
    usr->nick = name;
    return users[name] = usr;
}

void Network::removeUserChannel(NetUser *usr, NetChannel *chan)
{
    NetUserChannelList::Node *node;
    NetUserChannel *usrchan1 = NULL;
    NetUserChannel *usrchan2 = NULL;
    // Remove channel from user's list.
    for(node = usr->channels.GetFirst(); node != NULL; node = node->GetNext())
    {
        if(node->GetData()->getChannel() == chan)
        {
            usrchan1 = node->GetData();
            delete node;
            break;
        }
    }
    // Remove user from channel's list.
    for(node = chan->users.GetFirst(); node != NULL; node = node->GetNext())
    {
        if(node->GetData()->getUser() == usr)
        {
            if(node->GetData() != usrchan1)
                usrchan2 = node->GetData();
            delete node;
            break;
        }
    }
    // Delete the user-channel object(s).
    if(usrchan1 != NULL)
        delete usrchan1;
    if(usrchan2 != NULL)
        delete usrchan2;
}

void Network::addUserChannel(NetUserChannel *usrchan)
{
    usrchan->getUser()->channels.Append(usrchan);
    usrchan->getChannel()->users.Append(usrchan);
}

void Network::removeUser(NetUser *usr)
{
    if(usr == NULL)
        return;
    NetUserChannelList::Node *node;
    // Inform the manager that the user is about to be deleted (so that pointers to the object can be removed).
    if(mgr != NULL)
        mgr->receiveUserLeft(usr);
    // Remove user from all channels.
    while((node = usr->channels.GetFirst()) != NULL)
        removeUserChannel(node->GetData());
    // Delete the user.
    NetUserMap::iterator it = users.find(usr->nick);
    if(it != users.end() && it->second == usr)
        users.erase(it);
    delete usr;
#if 0
    // Clear information, but do not delete, since there may be pointers elsewhere to the object.
    usr->user.clear();
    usr->host.clear();
    usr->realname.clear();
    users.erase(usr->nick);
    // To avoid memory leaks, we need to keep the object in the map.
    // So we make an invalid nickname to use as the key.
    wxString key = usr->nick + wxT(" (defunct) ");
    while(users.find(key) != users.end())
        key += wxT("-");
    users[key] = usr;
    usr->nick = key;
#endif
}

void Network::removeUser(wxString nick)
{
    NetUserMap::iterator it = users.find(nick);
    if(it != users.end())
    {
        if(it->second == NULL)
            users.erase(it);
        else
            removeUser(it->second);
    }
}

void Network::changeUserNick(NetUser *usr, wxString newnick)
{
    if(usr == NULL || newnick.empty() || usr->nick == newnick)
        return;
    removeUser(newnick);
    NetUserMap::iterator it = users.find(usr->nick);
    if(it != users.end() && it->second == usr)
        users.erase(it);
    users[newnick] = usr;
    usr->nick = newnick;
}

void Network::changeUserNick(wxString oldnick, wxString newnick)
{
    if(newnick.empty() || oldnick == newnick)
        return;
    NetUserMap::iterator it = users.find(oldnick);
    if(it == users.end())
        return;
    removeUser(newnick);
    NetUser *usr = it->second;
    users[newnick] = usr;
    users.erase(it);
    if(usr != NULL)
        usr->nick = newnick;
}

bool Network::sendgamerequest(const wxString &gamid, const wxString &target, const wxString &message)
{
    return command(wxT("PRIVMSG ") + target + wxT(" :\x01" "MURDERSPOT ") + gamid + wxT(" ") + message + wxT("\x01"));
}

bool Network::respondgamerequest(const wxString &gamid, const wxString &target, const wxString &message)
{
    return TRUE;
//    return command(wxT("NOTICE ") + target + wxT(" :\x01" "MURDERSPOT ") + gamid + wxT(" ") + message + wxT("\x01"));
}

bool Network::chat(const wxString &target, const wxString &message)
{
    wxString msg;
    size_t pos1 = 0, pos2 = 0;
    while(pos1 < message.size() && pos2 < message.size())
    {
        pos2 = message.find_first_of(wxT("\r\n"), pos1);
        msg = message.substr(pos1, pos2 - pos1);
        pos1 = pos2 + 1;
        if(msg.empty())
            continue;
        if(target.empty())
        {
            if(!command(msg))
                return FALSE;
        }
        else
        {
            if(!command(wxT("PRIVMSG ") + target + wxT(" :") + msg))
                return FALSE;
        }
    }
    return TRUE;
}

bool Network::join(const wxString &channel, const wxString &key)
{
    if(channel.empty())
        return TRUE;
    return command(wxT("JOIN ") + channel + wxT(" :") + key);
}

bool Network::part(const wxString &channel, const wxString &msg)
{
    if(channel.empty())
        return TRUE;
    return command(wxT("PART ") + channel + wxT(" :") + msg);
}

BEGIN_EVENT_TABLE(NetClient, wxEvtHandler)
    EVT_SOCKET(-1, NetClient::OnSockEvent)
END_EVENT_TABLE()

NetClient::NetClient() : socket(NULL), ready(FALSE)
{
//printf("NetClient::NetClient()\n");
}

NetClient::~NetClient()
{
//printf("NetClient::~NetClient()\n");
    disconnect();
}

bool NetClient::connect(wxString server, unsigned short port, wxString nick, wxString username, wxString realname,
    wxString pass)
{
    disconnect();
    if(nick.empty())
        return FALSE;
    if(username.empty())
        username = nick;
    if(realname.empty())
        realname = nick;
    wxIPV4address addr;
    addr.Hostname(server);
    addr.Service(port);
    for(NetChannelMap::iterator cit = channels.begin(); cit != channels.end(); ++cit)
    {
        if(cit->second != NULL)
            delete cit->second;
    }
    channels.clear();
    pending.clear();
    immediatepending.clear();
    read.clear();
    ready = FALSE;
    for(NetUserMap::iterator uit = users.begin(); uit != users.end(); ++uit)
    {
        if(uit->second != NULL && uit->second != user)
            delete uit->second;
    }
    users.clear();
    if(user != NULL)
        delete user;
    user = users[nick] = new NetUser(this);
    user->nick = nick;
    user->user = username;
    user->host = wxGetFullHostName();
    user->realname = realname;
    socket = new wxSocketClient(wxSOCKET_NOWAIT);
    socket->SetEventHandler(*this);
    socket->SetNotify(wxSOCKET_INPUT_FLAG | wxSOCKET_OUTPUT_FLAG | wxSOCKET_CONNECTION_FLAG | wxSOCKET_LOST_FLAG);
    socket->Notify(TRUE);
    if(mgr != NULL)
        mgr->receiveNetStatus(this, wxT("Connecting..."));
    socket->Connect(addr, FALSE);
    if(!pass.empty())
        immediatecommand(wxT("PASS :") + pass);
    immediatecommand(wxT("NICK ") + nick);
    immediatecommand(wxT("USER ") + username + wxT(" ") + username + wxT(" ") + server + wxT(" :") + realname);
    return TRUE;
}

bool NetClient::disconnect(wxString quitmsg)
{
    if(socket != NULL)
    {
        if(socket->IsConnected())
            immediatecommand(wxT("QUIT :") + (quitmsg.empty() ? wxT("Leaving") : quitmsg));
        if(mgr != NULL)
        {
            mgr->receiveNetStatus(this, wxT("Disconnecting."), TRUE);
            mgr->receiveNetLost(this);
        }
        socket->Destroy();
        socket = NULL;
    }
    return TRUE;
}

bool NetClient::connected() const
{
    return socket != NULL && socket->IsConnected() && ready;
}

bool NetClient::command(const wxString &command)
{
    pending.append(command.mb_str(wxConvUTF8));
    pending.append("\r\n");
    sendData();
    return TRUE;
}

unsigned int NetClient::getMaxGameRequestLength(const wxString &gamid, const wxString &target) const
{
    // The full message will be:
    // :<nick>!<user>@<host> PRIVMSG <target> :@MURDERSPOT <gamid> <msg>@CL
    // (Where @ = '\1', C = '\r' and L = '\n')
    // This must be no longer than 512 bytes.
    // There are 28 characters there, aside from the variable bits.
    // Note that this will not quite work if there are wide characters involved...
    int len = 512 - 28;
    if(user == NULL)
    {
        len -= 75;
    }
    else
    {
        len -= strlen(user->nick.mb_str(wxConvUTF8));
        if(user->knowhostmask)
        {
            len -= strlen(user->user.mb_str(wxConvUTF8));
            len -= strlen(user->host.mb_str(wxConvUTF8));
        }
        else
        {
            len -= 50;
        }
    }
    len -= strlen(gamid.mb_str(wxConvUTF8));
    len -= strlen(target.mb_str(wxConvUTF8));
    if(len > 0)
        return len;
    return 0;
}

bool NetClient::immediatecommand(const wxString &command)
{
    immediatepending.append(command.mb_str(wxConvUTF8));
    immediatepending.append("\r\n");
    sendData();
    return TRUE;
}

void NetClient::sendData()
{
    if(socket != NULL && socket->IsConnected())
    {
        if(ready)
        {
            immediatepending.append(pending);
            pending.clear();
        }
        socket->Write(immediatepending.data(), immediatepending.size());
        immediatepending.erase(0, socket->LastCount());
    }
}

void NetClient::readData()
{
    if(socket != NULL && socket->IsConnected())
    {
        char buf[256];
        do
        {
            socket->Read(buf, 255);
            buf[socket->LastCount()] = 0;
            read.append(buf);
        } while(socket->LastCount() == 255);
    }
    size_t pos;
    wxString rawmsg;
    wxString msg;
    wxString msg2;
    wxString cmd;
    wxString arg;
    wxString rec;
    NetUser *sender;
    NetUser *affected;
    NetChannel *target;
    NetUserChannelList::Node *node;
    int level;
    while(1)
    {
        // Although "\r\n" is the correct line termination, we accept either character on its own.
        // This was added for compatibility with irssi-proxy, which the author wanted to use.
        // A correct "\r\n" (or even "\n\r") will still work, as blank lines are ignored.
        pos = read.find_first_of("\r\n");
        if(pos == std::string::npos)
            break;
        msg = rawmsg = wxString(read.substr(0, pos).c_str(), wxConvUTF8);
        read.erase(0, pos + 1);
        if(msg.empty())
            continue;
        cmd = nextIRCparam(msg, FALSE);
        if(cmd.empty())
            continue;
        if(cmd.at(0) == wxT(':'))
        {
            // Sender specified - get details.
            pos = cmd.find(wxT("!"), 1);
            if(pos < cmd.size())
            {
                sender = fetchUser(cmd.substr(1, pos - 1));
                cmd.erase(0, pos + 1);
                pos = cmd.find(wxT("@"));
                if(pos < wxString::npos)
                {
                    sender->user = cmd.substr(0, pos);
                    sender->host = cmd.substr(pos + 1);
                }
                else
                {
                    sender->user = cmd;
                    sender->host.clear();
                }
                sender->knowhostmask = TRUE;
            }
            else if(cmd.find(wxT("."), 1) < cmd.size())
            {
                sender = NULL;
            }
            else
            {
                sender = fetchUser(cmd);
            }
            cmd = nextIRCparam(msg, FALSE);
        }
        else
        {
            sender = NULL;
        }
//        if(cmd == wxT("001") || cmd == wxT("353") || cmd == wxT("331") || cmd == wxT("332") || cmd == wxT("324") || cmd == wxT("329"))
        if(cmd.size() == 3 && cmd[0] >= '0' && cmd[0] <= '9' && cmd[1] >= '0' && cmd[1] <= '9' && cmd[2] >= '0' && cmd[2] <= '9')
        {
            // Numeric response code - The first parameter should be our nick.
            arg = nextIRCparam(msg);
            if(arg != user->nick && !arg.empty() && arg[0] != wxT('*'))
            {
                users.erase(user->nick);
                user->nick = arg;
                removeUser(arg);
                users[arg] = user;
                if(mgr != NULL)
                    mgr->changeNick(this, arg);
            }
        }
        else
        {
            cmd.MakeUpper();
        }
        if(cmd == wxT("001"))
        {
            // Registered with server.
            ready = TRUE;
            if(mgr != NULL)
            {
                mgr->receiveNetStatus(this, rawmsg);
                mgr->changeNick(this, user->nick);
            }
#if 0
            // Send a message to ourselves, in the hope of finding our own host mask.
            immediatecommand(wxT("PRIVMSG ") + user->nick + wxT(" :\x1" "MURDERSPOT NOOP\x1"));
#endif
            sendData();
        }
        else if(cmd == wxT("433"))
        {
            // Nickname already in use.
            // Try another if not yet registered.
            if(!ready)
            {
                users.erase(user->nick);
                user->nick.insert(0, wxT("_"));
                removeUser(user->nick);
                users[user->nick] = user;
                immediatecommand(wxT("NICK ") + user->nick);
            }
            if(mgr != NULL)
                mgr->receiveNetStatus(this, rawmsg);
        }
        else if(cmd == wxT("PING"))
        {
            // Reply to PING with PONG.
            immediatecommand(wxT("PONG ") + msg);
        }
        else if(cmd == wxT("353"))
        {
            // NAMES reply - update users on channel.
#if 0
            arg = nextIRCparam(msg); // Should be our nick.
            if(arg != user->nick)
                continue;
#endif
            arg = nextIRCparam(msg); // Should be = character.
            if(arg != wxT("="))
                continue;
            rec = nextIRCparam(msg); // Channel name.
            if(rec.empty() || rec.at(0) != wxT('#') || msg.empty())
                continue;
            target = fetchChannel(rec);
            if(msg.at(0) == wxT(':'))
                msg.erase(0, 1);
            if(mgr != NULL)
                mgr->receiveChat(this, wxEmptyString, rec, wxT("NAMES"), msg);
            arg = nextIRCparam(msg);
            while(!arg.empty())
            {
                level = NetUserChannel::NORMAL;
                while(1)
                {
                    if(arg[0] == wxT('+'))
                    {
                        if(level < NetUserChannel::VOICE)
                            level = NetUserChannel::VOICE;
                    }
                    else if(arg[0] == wxT('%'))
                    {
                        if(level < NetUserChannel::HALFOP)
                            level = NetUserChannel::HALFOP;
                    }
                    else if(arg[0] == wxT('@'))
                    {
                        if(level < NetUserChannel::OP)
                            level = NetUserChannel::OP;
                    }
                    else if(arg[0] == wxT('&'))
                    {
                        if(level < NetUserChannel::ADMIN)
                            level = NetUserChannel::ADMIN;
                    }
                    else if(arg[0] == wxT('~') || arg[0] == wxT('.'))
                    {
                        if(level < NetUserChannel::OWNER)
                            level = NetUserChannel::OWNER;
                    }
                    else
                    {
                        break;
                    }
                    arg.erase(0, 1);
                }
                sender = fetchUser(arg);
                removeUserChannel(sender, target);
                addUserChannel(sender, target, level);
                arg = nextIRCparam(msg);
            }
        }
        else if(cmd == wxT("366"))
        {
            // End of NAMES reply - discard.
        }
        else if(cmd == wxT("331"))
        {
            // No topic is set.
            rec = nextIRCparam(msg);
            if(mgr != NULL)
                mgr->receiveChat(this, wxEmptyString, rec, wxT("TOPIC"), wxEmptyString);
        }
        else if(cmd == wxT("332") || cmd == wxT("TOPIC"))
        {
            // Topic notification.
            rec = nextIRCparam(msg);
            if(sender == NULL)
                arg = wxEmptyString;
            else
                arg = sender->nick;
            if(mgr != NULL)
                mgr->receiveChat(this, arg, rec, wxT("TOPIC"), nextIRCparam(msg));
        }
        else if(cmd == wxT("333"))
        {
            // Topic change date notification.
            // For now, this is discarded.
            // TODO: Send this to the window, and have it display it properly.
        }
        else if(cmd == wxT("324"))
        {
            // Channel mode notification.
            rec = nextIRCparam(msg);
            if(mgr != NULL)
                mgr->receiveChat(this, wxEmptyString, rec, wxT("MODE"), msg);
        }
        else if(cmd == wxT("329"))
        {
            // Channel creation date.
            // TODO: As for 333
        }
        else if(cmd == wxT("321"))
        {
            // The beginning of a LIST response.
            if(mgr != NULL)
                mgr->receiveChat(this, wxEmptyString, wxEmptyString, wxT("LISTSTART"), msg);
        }
        else if(cmd == wxT("322"))
        {
            // A channel in a LIST response.
            if(mgr != NULL)
                mgr->receiveChat(this, wxEmptyString, wxEmptyString, wxT("LIST"), msg);
        }
        else if(cmd == wxT("323"))
        {
            // The end of a LIST response.
            if(mgr != NULL)
                mgr->receiveChat(this, wxEmptyString, wxEmptyString, wxT("LISTEND"), msg);
        }
        else if(sender != NULL)
        {
            if(cmd == wxT("JOIN"))
            {
                rec = nextIRCparam(msg);
                if(rec.empty() || rec.at(0) != wxT('#'))
                    continue;
                target = fetchChannel(rec);
                removeUserChannel(sender, target);
                addUserChannel(sender, target);
                if(mgr != NULL)
                {
                    if(sender != user)
                        mgr->receiveGameRequest(this, rec, sender, wxT("JOIN"), wxT(""));
                    mgr->receiveChat(this, sender->nick, rec, cmd, nextIRCparam(msg));
                }
            }
            else if(cmd == wxT("PART"))
            {
                rec = nextIRCparam(msg);
                if(rec.empty() || rec.at(0) != wxT('#'))
                    continue;
                target = fetchChannel(rec);
                removeUserChannel(sender, target);
                if(mgr != NULL)
                {
                    if(sender == user)
                        mgr->receiveGameLeft(this, rec);
                    else
                        mgr->receiveGameRequest(this, rec, sender, wxT("ENDGAME"), wxT(""));
                    mgr->receiveChat(this, sender->nick, rec, cmd, nextIRCparam(msg));
                }
            }
            else if(cmd == wxT("KICK"))
            {
                rec = nextIRCparam(msg);
                if(rec.empty() || rec.at(0) != wxT('#'))
                    continue;
                target = fetchChannel(rec);
                arg = nextIRCparam(msg);
                affected = fetchUser(arg);
                removeUserChannel(affected, target);
                if(mgr != NULL)
                {
                    if(affected == user)
                        mgr->receiveGameLeft(this, rec);
                    else
                        mgr->receiveGameRequest(this, rec, sender, wxT("ENDGAME"), wxT(""));
                    mgr->receiveChat(this, sender->nick, rec, cmd, arg + wxT(" ") + nextIRCparam(msg));
                }
            }
            else if(cmd == wxT("QUIT"))
            {
                arg = nextIRCparam(msg);
                while((node = sender->channels.GetFirst()) != NULL)
                {
                    rec = node->GetData()->getChannel()->name;
                    removeUserChannel(node->GetData());
                    if(mgr != NULL)
                    {
                        if(sender != user)
                            mgr->receiveGameRequest(this, rec, sender, wxT("ENDGAME"), wxT(""));
                        mgr->receiveChat(this, sender->nick, rec, cmd, arg);
                    }
                }
                if(mgr != NULL)
                    mgr->receiveChat(this, sender->nick, user->nick, cmd, arg);
                removeUser(sender);
                sender = NULL;
            }
            else if(cmd == wxT("NICK"))
            {
                arg = nextIRCparam(msg);
                if(arg != sender->nick)
                {
                    rec = sender->nick;
                    removeUser(arg);
                    users.erase(sender->nick);
                    sender->nick = arg;
                    users[arg] = sender;
                    if(mgr != NULL)
                    {
                        mgr->receiveChat(this, rec, user->nick, cmd, arg);
                        for(node = sender->channels.GetFirst(); node != NULL; node = node->GetNext())
                            mgr->receiveChat(this, rec, node->GetData()->getChannel()->name, cmd, arg);
                        if(sender->nick == user->nick)
                            mgr->changeNick(this, arg);
                    }
                }
            }
            else if(cmd == wxT("MODE"))
            {
                // Mode change - identify the channel, and send to the window.
                rec = nextIRCparam(msg);
                if(mgr != NULL)
                    mgr->receiveChat(this, sender->nick, rec, wxT("MODE"), msg);
            }
            else if(cmd == wxT("PRIVMSG") || cmd == wxT("NOTICE"))
            {
                // A message - first find the target channel, if any.
                rec = nextIRCparam(msg);
                if(rec.empty() || rec.at(0) != wxT('#'))
                    target = NULL;
                else
                    target = fetchChannel(rec);
                // Get message content.
                msg = nextIRCparam(msg);
                // Handle CTCP messages.
                if(msg[0] == '\x01' && msg[msg.size() - 1] == '\x01')
                {
                    msg.erase(0, 1);
                    msg.erase(msg.size() - 1, 1);
                    arg = nextIRCparam(msg, FALSE);
                    arg.MakeUpper();
                    if(cmd == wxT("PRIVMSG"))
                    {
                        if(arg == wxT("PING"))
                            msg2 = wxT("PING ") + msg;
                        else if(arg == wxT("VERSION"))
                            msg2 = wxT("VERSION " PACKAGE_STRING " ") + wxGetOsDescription();
                        else
                            msg2.clear();
                        if(mgr != NULL)
                        {
                            if(arg == wxT("MURDERSPOT"))
                            {
                                rec = nextIRCparam(msg, FALSE);
                                rec.MakeLower();
                                arg = nextIRCparam(msg, FALSE);
                                arg.MakeUpper();
                                mgr->receiveGameRequest(this, rec, sender, arg, msg);
                            }
                            else if(arg == wxT("ACTION"))
                            {
                                mgr->receiveChat(this, sender->nick, rec, wxT("ACTION"), msg);
                            }
                            else
                            {
                                mgr->receiveChat(this, sender->nick, rec, wxT("CTCP"), arg + wxT(" ") + msg);
                            }
                            if(!msg2.empty())
                                mgr->receiveChat(this, user->nick, sender->nick, wxT("NCTCP"), msg2);
                        }
                        if(!msg2.empty())
                            command(wxT("NOTICE ") + sender->nick + wxT(" :\x01") + msg2 + wxT("\x01"));
                    }
                    else if(mgr != NULL)
                        mgr->receiveChat(this, sender->nick, rec, wxT("NCTCP"), arg + wxT(" ") + msg);
                }
                else if(mgr != NULL)
                    mgr->receiveChat(this, sender->nick, rec, cmd, msg);
            }
            else if(mgr != NULL)
                mgr->receiveNetStatus(this, rawmsg);
        }
        else if(mgr != NULL)
            mgr->receiveNetStatus(this, rawmsg);
    }
}

void NetClient::OnSockEvent(wxSocketEvent &event)
{
    if(socket == NULL || event.GetSocket() != socket)
    {
        event.Skip();
    }
    else
    {
        switch(event.GetSocketEvent())
        {
        case wxSOCKET_INPUT:
            readData();
            break;
        case wxSOCKET_OUTPUT:
            sendData();
            break;
        case wxSOCKET_CONNECTION:
            if(mgr != NULL)
                mgr->receiveNetStatus(this, wxT("Connection established."), TRUE);
            sendData();
            readData();
            break;
        case wxSOCKET_LOST:
            socket->Close();
            if(mgr != NULL)
            {
                mgr->receiveNetStatus(this, wxT("Lost connection."), TRUE);
                mgr->receiveNetLost(this);
            }
            break;
        default:
            event.Skip();
        }
    }
}
