// MurderSpot - a free game based on Clue/Cluedo
// Copyright 2005-2009 Christopher Elsby <chrise@chrise.me.uk>
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef MURDERSPOT_NETWORK_H
#define MURDERSPOT_NETWORK_H

#include <wx/string.h>
#include <wx/hashmap.h>
#include <wx/list.h>
#include <wx/event.h>
#include <string> // If wxWidgets won't supply a dynamic string class always using char...

class Manager;
class NetUser;
class NetChannel;
class NetUserChannel;
class Network;
class wxSocketClient;
class wxSocketEvent;

WX_DECLARE_STRING_HASH_MAP(NetUser*, NetUserMap);
WX_DECLARE_STRING_HASH_MAP(NetChannel*, NetChannelMap);
WX_DECLARE_LIST(NetUserChannel, NetUserChannelList);

class NetUser
{
public:
    NetUser(Network *net) : knowhostmask(FALSE), network(net) { }
    Network *getNetwork() const {return network;}
//    const Network *getNetwork() const {return network;}
    void setNetwork(Network *net) {network = net;}
    unsigned int getMaxGameRequestLength(const wxString &gamid) const;
    bool sendgamerequest(const wxString &gamid, const wxString &message) const;
    bool respondgamerequest(const wxString &gamid, const wxString &message) const;
    bool chat(const wxString &message) const;
    NetUserChannel *getChannelData(const NetChannel *chan) const;
//    bool canHost(const NetChannel *chan) const;
    wxString getHostMask() const;
    wxString nick;
    wxString user;
    wxString host;
    wxString realname;
    bool knowhostmask;
    NetUserChannelList channels;
private:
    Network *network;
};

class NetChannel
{
public:
    NetChannel(Network *net) : network(net) { }
    Network *getNetwork() const {return network;}
    void setNetwork(Network *net);
    unsigned int getMaxGameRequestLength(const wxString &gamid) const;
    bool sendgamerequest(const wxString &gamid, const wxString &message) const;
    bool respondgamerequest(const wxString &gamid, const wxString &message) const;
    bool chat(const wxString &message) const;
    bool join() const;
    bool part(const wxString &msg = wxEmptyString) const;
    NetUserChannel *getUserData(const NetUser *usr) const;
    const NetUser *getUser() const;
    const NetUser *getUser(const wxString &name) const;
    wxString name;
    wxString topic;
    NetUserChannelList users;
private:
    Network *network;
};


const wxChar nickprefixes[] = wxT(" +%@&~");
class NetUserChannel
{
public:
    enum { NORMAL, VOICE, HALFOP, OP, ADMIN, OWNER };
    NetUserChannel(NetUser *usr = NULL, NetChannel *chan = NULL, int lev = NORMAL) : user(usr), channel(chan), level(lev) { }
    NetUser *getUser() {return user;}
    const NetUser *getUser() const {return user;}
    void setUser(NetUser *usr) {user = usr;}
    NetChannel *getChannel() {return channel;}
    const NetChannel *getChannel() const {return channel;}
    void setChannel(NetChannel *chan) {channel = chan;}
    Network *getNetwork() const {if(user != NULL) return user->getNetwork(); if(channel != NULL) return channel->getNetwork(); return NULL;}
    int getLevel() const {return level;}
    wxString getPrefix() const {return nickprefixes[level];}
    void setLevel(int lev) {level = lev;}
    bool hasLevel(int lev) const {return level >= lev;}
//    bool canHost() const {return level >= HALFOP;}
private:
    NetUser *user;
    NetChannel *channel;
    int level;
};

class Network
{
public:
    Network() : mgr(NULL), user(NULL) { }
    virtual ~Network();
    virtual bool disconnect(wxString quitmsg = wxEmptyString) = 0;
    virtual bool connected() const = 0;
    virtual bool command(const wxString &command) = 0;
    virtual bool immediatecommand(const wxString &command) = 0;
    virtual unsigned int getMaxGameRequestLength(const wxString &gamid, const wxString &target) const = 0;
    const NetChannel *getChannel(wxString name) const;
    bool sendgamerequest(const wxString &gamid, const wxString &target, const wxString &message);
    bool respondgamerequest(const wxString &gamid, const wxString &sender, const wxString &message);
    bool chat(const wxString &target, const wxString &message);
    bool join(const wxString &channel, const wxString &key = wxEmptyString);
    bool part(const wxString &channel, const wxString &msg = wxEmptyString);
    wxString getNick() const {if(user == NULL) return wxEmptyString; return user->nick;}
    const NetUser *getUser() const {return user;}
    const NetUser *getUser(wxString name) const;
    Manager *mgr;
protected:
    NetChannel *fetchChannel(wxString name);
    NetUser *fetchUser() {return user;}
    NetUser *fetchUser(wxString name);
    void removeUserChannel(NetUserChannel *usrchan) {removeUserChannel(usrchan->getUser(), usrchan->getChannel());}
    void removeUserChannel(NetUser *usr, NetChannel *chan);
    void addUserChannel(NetUserChannel *usrchan);
    void addUserChannel(NetUser *usr, NetChannel *chan, int level = NetUserChannel::NORMAL)
        {addUserChannel(new NetUserChannel(usr, chan, level));}
    void removeUser(NetUser *usr);
    void removeUser(wxString nick);
    void changeUserNick(NetUser *usr, wxString newnick);
    void changeUserNick(wxString oldnick, wxString newnick);
    NetUser *user;
    NetUserMap users;
    NetChannelMap channels;
};

inline unsigned int NetUser::getMaxGameRequestLength(const wxString &gamid) const
{if(network == NULL) return 0; return network->getMaxGameRequestLength(gamid, nick);}
inline bool NetUser::sendgamerequest(const wxString &gamid, const wxString &message) const
{return network != NULL && network->sendgamerequest(gamid, nick, message);}
inline bool NetUser::respondgamerequest(const wxString &gamid, const wxString &message) const
{return network != NULL && network->respondgamerequest(gamid, nick, message);}
inline bool NetUser::chat(const wxString &message) const {return network != NULL && network->chat(nick, message);}

inline unsigned int NetChannel::getMaxGameRequestLength(const wxString &gamid) const
{if(network == NULL) return 0; return network->getMaxGameRequestLength(gamid, name);}
inline bool NetChannel::sendgamerequest(const wxString &gamid, const wxString &message) const
{return network != NULL && network->sendgamerequest(gamid, name, message);}
inline bool NetChannel::respondgamerequest(const wxString &gamid, const wxString &message) const
{return network != NULL && network->respondgamerequest(gamid, name, message);}
inline bool NetChannel::chat(const wxString &message) const {return network != NULL && network->chat(name, message);}
inline bool NetChannel::join() const {return network != NULL && network->join(name);}
inline bool NetChannel::part(const wxString &msg) const {return network != NULL && network->part(name, msg);}
inline const NetUser *NetChannel::getUser() const {if(network == NULL) return NULL; return network->getUser();}
inline const NetUser *NetChannel::getUser(const wxString &name) const {if(network == NULL) return NULL; return network->getUser(name);}

class NetClient : public Network, public wxEvtHandler
{
public:
    NetClient();
    virtual ~NetClient();
    bool connect(wxString server, unsigned short port, wxString nick, wxString username, wxString realname, wxString pass);
    virtual bool disconnect(wxString quitmsg = wxEmptyString);
    virtual bool connected() const;
    virtual bool command(const wxString &command);
    virtual unsigned int getMaxGameRequestLength(const wxString &gamid, const wxString &target) const;
    virtual bool immediatecommand(const wxString &command);
private:
    void sendData();
    void readData();
    void OnSockEvent(wxSocketEvent &event);
    wxSocketClient *socket;
    std::string immediatepending;
    std::string pending;
    std::string read;
    bool ready;
    DECLARE_EVENT_TABLE()
};

#endif // !defined(MURDERSPOT_NETWORK_H)
